#
# pjsua Python GUI Demo
#
# Copyright (C)2013 Teluu Inc. (http://www.teluu.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
#
import sys
if sys.version_info[0] >= 3: # Python 3
    import tkinter as tk
    from tkinter import ttk
    from tkinter import messagebox as msgbox
else:
    import Tkinter as tk
    import tkMessageBox as msgbox
    import ttk

import random
import pjsua2 as pj
import endpoint
import application

# Buddy class
class Buddy(pj.Buddy):
    """
    High level Python Buddy object, derived from pjsua2's Buddy object.
    """
    def __init__(self, app):
        pj.Buddy.__init__(self)
        self.app = app
        self.randId = random.randint(1, 9999)
        self.cfg = None
        self.account = None

    def statusText(self):
        bi = self.getInfo()
        status = ''
        if bi.subState == pj.PJSIP_EVSUB_STATE_ACTIVE:
            if bi.presStatus.status == pj.PJSUA_BUDDY_STATUS_ONLINE:
                status = bi.presStatus.statusText
                if not status:
                    status = 'Online'
            elif bi.presStatus.status == pj.PJSUA_BUDDY_STATUS_OFFLINE:
                status = 'Offline'
            else:
                status = 'Unknown'
        return status

    def onBuddyState(self):
        self.app.updateBuddy(self)

class SettingDialog(tk.Toplevel):
    """
    This implements buddy settings dialog to manipulate buddy settings.
    """
    def __init__(self, parent, cfg):
        tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        self.parent = parent
        self.geometry("+100+100")
        self.title('Buddy settings')

        self.frm = ttk.Frame(self)
        self.frm.pack(expand='yes', fill='both')

        self.isOk = False
        self.cfg = cfg

        self.createWidgets()

    def doModal(self):
        if self.parent:
            self.parent.wait_window(self)
        else:
            self.wait_window(self)
        return self.isOk

    def createWidgets(self):
        # The notebook
        self.frm.rowconfigure(0, weight=1)
        self.frm.rowconfigure(1, weight=0)
        self.frm.columnconfigure(0, weight=1)
        self.frm.columnconfigure(1, weight=1)
        self.wTab = ttk.Notebook(self.frm)
        self.wTab.grid(column=0, row=0, columnspan=2, padx=5, pady=5, sticky=tk.N+tk.S+tk.W+tk.E)

        # Main buttons
        btnOk = ttk.Button(self.frm, text='Ok', command=self.onOk)
        btnOk.grid(column=0, row=1, sticky=tk.E, padx=20, pady=10)
        btnCancel = ttk.Button(self.frm, text='Cancel', command=self.onCancel)
        btnCancel.grid(column=1, row=1, sticky=tk.W, padx=20, pady=10)

        # Tabs
        self.createBasicTab()

    def createBasicTab(self):
        # Prepare the variables to set/receive values from GUI
        self.cfgUri = tk.StringVar()
        self.cfgUri.set( self.cfg.uri )
        self.cfgSubscribe = tk.BooleanVar(value = self.cfg.subscribe)

        # Build the tab page
        frm = ttk.Frame(self.frm)
        frm.columnconfigure(0, weight=1)
        frm.columnconfigure(1, weight=2)
        row = 0
        ttk.Label(frm, text='URI:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Entry(frm, textvariable=self.cfgUri, width=40).grid(row=row, column=1, sticky=tk.W+tk.E, padx=6)
        row += 1
        ttk.Checkbutton(frm, text='Subscribe presence', variable=self.cfgSubscribe).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)

        self.wTab.add(frm, text='Basic Settings')


    def onOk(self):
        # Check basic settings
        errors = "";
        if self.cfgUri.get():
            if not endpoint.validateSipUri(self.cfgUri.get()):
                errors += "Invalid Buddy URI: '%s'\n" % (self.cfgUri.get())

        if errors:
            msgbox.showerror("Error detected:", errors)
            return

        # Basic settings
        self.cfg.uri = self.cfgUri.get()
        self.cfg.subscribe = self.cfgSubscribe.get()

        self.isOk = True
        self.destroy()

    def onCancel(self):
        self.destroy()


if __name__ == '__main__':
    application.main()



#
from inc_cfg import *

# ICE mismatch
test_param = TestParam(
		"Callee=use ICE & SRTP, caller=use ICE & DTLS-SRTP",
		[
			InstanceParam("callee", "--null-audio --use-ice --max-calls=1 --use-srtp=2 --srtp-secure=0"),
			InstanceParam("caller", "--null-audio --use-ice --max-calls=1 --use-srtp=2 --srtp-secure=0 --srtp-keying=1")
		]
		)



#
# pjsua Python GUI Demo
#
# Copyright (C)2013 Teluu Inc. (http://www.teluu.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
#
import sys
if sys.version_info[0] >= 3: # Python 3
    import tkinter as tk
    from tkinter import ttk
    from tkinter import messagebox as msgbox
else:
    import Tkinter as tk
    import tkMessageBox as msgbox
    import ttk

import pjsua2 as pj
import endpoint
import application

class Dialog(tk.Toplevel):
    """
    This implements account settings dialog to manipulate account settings.
    """
    def __init__(self, parent, cfg):
        tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        self.parent = parent
        self.geometry("+100+100")
        self.title('Account settings')

        self.frm = ttk.Frame(self)
        self.frm.pack(expand='yes', fill='both')

        self.isOk = False
        self.cfg = cfg

        self.createWidgets()

    def doModal(self):
        if self.parent:
            self.parent.wait_window(self)
        else:
            self.wait_window(self)
        return self.isOk

    def createWidgets(self):
        # The notebook
        self.frm.rowconfigure(0, weight=1)
        self.frm.rowconfigure(1, weight=0)
        self.frm.columnconfigure(0, weight=1)
        self.frm.columnconfigure(1, weight=1)
        self.wTab = ttk.Notebook(self.frm)
        self.wTab.grid(column=0, row=0, columnspan=2, padx=10, pady=10, ipadx=20, ipady=20, sticky=tk.N+tk.S+tk.W+tk.E)

        # Main buttons
        btnOk = ttk.Button(self.frm, text='Ok', command=self.onOk)
        btnOk.grid(column=0, row=1, sticky=tk.E, padx=20, pady=10)
        btnCancel = ttk.Button(self.frm, text='Cancel', command=self.onCancel)
        btnCancel.grid(column=1, row=1, sticky=tk.W, padx=20, pady=10)

        # Tabs
        self.createBasicTab()
        self.createSipTab()
        self.createMediaTab()
        self.createMediaNatTab()

    def createBasicTab(self):
        # Prepare the variables to set/receive values from GUI
        self.cfgPriority = tk.IntVar(value=self.cfg.priority)
        self.cfgAccId = tk.StringVar(value=self.cfg.idUri)
        self.cfgRegistrar = tk.StringVar(value=self.cfg.regConfig.registrarUri)
        self.cfgRegisterOnAdd = tk.BooleanVar(value=self.cfg.regConfig.registerOnAdd)
        self.cfgUsername = tk.StringVar()
        self.cfgPassword = tk.StringVar()
        if len(self.cfg.sipConfig.authCreds):
            self.cfgUsername.set( self.cfg.sipConfig.authCreds[0].username )
            self.cfgPassword.set( self.cfg.sipConfig.authCreds[0].data )
        self.cfgProxy = tk.StringVar()
        if len(self.cfg.sipConfig.proxies):
            self.cfgProxy.set( self.cfg.sipConfig.proxies[0] )

        # Build the tab page
        frm = ttk.Frame(self.frm)
        frm.columnconfigure(0, weight=1)
        frm.columnconfigure(1, weight=2)
        row = 0
        ttk.Label(frm, text='Priority:').grid(row=row, column=0, sticky=tk.E, pady=2)
        tk.Spinbox(frm, from_=0, to=9, textvariable=self.cfgPriority, width=2).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='ID (URI):').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Entry(frm, textvariable=self.cfgAccId, width=32).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='Registrar URI:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Entry(frm, textvariable=self.cfgRegistrar, width=32).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Checkbutton(frm, text='Register on add', variable=self.cfgRegisterOnAdd).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Label(frm, text='Optional proxy URI:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Entry(frm, textvariable=self.cfgProxy, width=32).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='Auth username:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Entry(frm, textvariable=self.cfgUsername, width=16).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='Password:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Entry(frm, textvariable=self.cfgPassword, show='*', width=16).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)

        self.wTab.add(frm, text='Basic Settings')


    def createSipTab(self):
        # Prepare the variables to set/receive values from GUI
        self.cfgPrackUse 	= tk.IntVar(value=self.cfg.callConfig.prackUse)
        self.cfgTimerUse 	= tk.IntVar(value=self.cfg.callConfig.timerUse)
        self.cfgTimerExpires 	= tk.IntVar(value=self.cfg.callConfig.timerSessExpiresSec)
        self.cfgPublish 	= tk.BooleanVar(value=self.cfg.presConfig.publishEnabled)
        self.cfgMwiEnabled 	= tk.BooleanVar(value=self.cfg.mwiConfig.enabled)
        self.cfgEnableContactRewrite = tk.BooleanVar(value=self.cfg.natConfig.contactRewriteUse != 0)
        self.cfgEnableViaRewrite = tk.BooleanVar(value=self.cfg.natConfig.viaRewriteUse != 0)
        self.cfgEnableSdpRewrite = tk.BooleanVar(value=self.cfg.natConfig.sdpNatRewriteUse != 0)
        self.cfgEnableSipOutbound = tk.BooleanVar(value=self.cfg.natConfig.sipOutboundUse != 0)
        self.cfgKaInterval 	= tk.IntVar(value=self.cfg.natConfig.udpKaIntervalSec)

        # Build the tab page
        frm = ttk.Frame(self.frm)
        frm.columnconfigure(0, weight=1)
        frm.columnconfigure(1, weight=2)
        row = 0
        ttk.Label(frm, text='100rel/PRACK:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Radiobutton(frm, text='Only offer PRACK', value=pj.PJSUA_100REL_NOT_USED, variable=self.cfgPrackUse).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Offer and use if remote supports', value=pj.PJSUA_100REL_OPTIONAL, variable=self.cfgPrackUse).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Required', value=pj.PJSUA_100REL_MANDATORY, variable=self.cfgPrackUse).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='Session Timer:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Radiobutton(frm, text='Not offered', value=pj.PJSUA_SIP_TIMER_INACTIVE, variable=self.cfgTimerUse).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Optional', value=pj.PJSUA_SIP_TIMER_OPTIONAL, variable=self.cfgTimerUse).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Required', value=pj.PJSUA_SIP_TIMER_REQUIRED, variable=self.cfgTimerUse).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text="Always use", value=pj.PJSUA_SIP_TIMER_ALWAYS, variable=self.cfgTimerUse).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='Session Timer Expiration:').grid(row=row, column=0, sticky=tk.E, pady=2)
        tk.Spinbox(frm, from_=90, to=7200, textvariable=self.cfgTimerExpires, width=5).grid(row=row, column=1, sticky=tk.W, padx=6)
        ttk.Label(frm, text='(seconds)').grid(row=row, column=1, sticky=tk.E)
        row += 1
        ttk.Label(frm, text='Presence:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Checkbutton(frm, text='Enable PUBLISH', variable=self.cfgPublish).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Label(frm, text='Message Waiting Indication:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Checkbutton(frm, text='Enable MWI', variable=self.cfgMwiEnabled).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Label(frm, text='NAT Traversal:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Checkbutton(frm, text='Enable Contact Rewrite', variable=self.cfgEnableContactRewrite).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Checkbutton(frm, text='Enable Via Rewrite', variable=self.cfgEnableViaRewrite).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Checkbutton(frm, text='Enable SDP IP Address Rewrite', variable=self.cfgEnableSdpRewrite).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Checkbutton(frm, text='Enable SIP Outbound Extension', variable=self.cfgEnableSipOutbound).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Label(frm, text='UDP Keep-Alive Interval:').grid(row=row, column=0, sticky=tk.E, pady=2)
        tk.Spinbox(frm, from_=0, to=3600, textvariable=self.cfgKaInterval, width=5).grid(row=row, column=1, sticky=tk.W, padx=6)
        ttk.Label(frm, text='(seconds) Zero to disable.').grid(row=row, column=1, sticky=tk.E)


        self.wTab.add(frm, text='SIP Features')

    def createMediaTab(self):
        # Prepare the variables to set/receive values from GUI
        self.cfgMedPort = tk.IntVar(value=self.cfg.mediaConfig.transportConfig.port)
        self.cfgMedPortRange = tk.IntVar(value=self.cfg.mediaConfig.transportConfig.portRange)
        self.cfgMedLockCodec = tk.BooleanVar(value=self.cfg.mediaConfig.lockCodecEnabled)
        self.cfgMedSrtp = tk.IntVar(value=self.cfg.mediaConfig.srtpUse)
        self.cfgMedSrtpSecure = tk.IntVar(value=self.cfg.mediaConfig.srtpSecureSignaling)
        self.cfgMedIpv6 = tk.BooleanVar(value=self.cfg.mediaConfig.ipv6Use==pj.PJSUA_IPV6_ENABLED)

        # Build the tab page
        frm = ttk.Frame(self.frm)
        frm.columnconfigure(0, weight=1)
        frm.columnconfigure(1, weight=21)
        row = 0
        ttk.Label(frm, text='Secure RTP (SRTP):').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Radiobutton(frm, text='Disable', value=pj.PJMEDIA_SRTP_DISABLED, variable=self.cfgMedSrtp).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Mandatory', value=pj.PJMEDIA_SRTP_MANDATORY, variable=self.cfgMedSrtp).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Optional (non-standard)', value=pj.PJMEDIA_SRTP_OPTIONAL, variable=self.cfgMedSrtp).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='SRTP signaling:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Radiobutton(frm, text='Does not require secure signaling', value=0, variable=self.cfgMedSrtpSecure).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Require secure next hop (TLS)', value=1, variable=self.cfgMedSrtpSecure).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Require secure end-to-end (SIPS)', value=2, variable=self.cfgMedSrtpSecure).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='RTP transport start port:').grid(row=row, column=0, sticky=tk.E, pady=2)
        tk.Spinbox(frm, from_=0, to=65535, textvariable=self.cfgMedPort, width=5).grid(row=row, column=1, sticky=tk.W, padx=6)
        ttk.Label(frm, text='(0: any)').grid(row=row, column=1, sticky=tk.E, pady=2)
        row += 1
        ttk.Label(frm, text='Port range:').grid(row=row, column=0, sticky=tk.E, pady=2)
        tk.Spinbox(frm, from_=0, to=65535, textvariable=self.cfgMedPortRange, width=5).grid(row=row, column=1, sticky=tk.W, padx=6)
        ttk.Label(frm, text='(0: not limited)').grid(row=row, column=1, sticky=tk.E, pady=2)
        row += 1
        ttk.Label(frm, text='Lock codec:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Checkbutton(frm, text='Enable', variable=self.cfgMedLockCodec).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Label(frm, text='Use IPv6:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Checkbutton(frm, text='Yes', variable=self.cfgMedIpv6).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)

        self.wTab.add(frm, text='Media settings')

    def createMediaNatTab(self):
        # Prepare the variables to set/receive values from GUI
        self.cfgSipUseStun = tk.IntVar(value = self.cfg.natConfig.sipStunUse)
        self.cfgMediaUseStun = tk.IntVar(value = self.cfg.natConfig.mediaStunUse)
        self.cfgIceEnabled = tk.BooleanVar(value = self.cfg.natConfig.iceEnabled)
        self.cfgIceAggressive = tk.BooleanVar(value = self.cfg.natConfig.iceAggressiveNomination)
        self.cfgAlwaysUpdate = tk.BooleanVar(value = True if self.cfg.natConfig.iceAlwaysUpdate else False)
        self.cfgIceNoHostCands = tk.BooleanVar(value = True if self.cfg.natConfig.iceMaxHostCands == 0 else False)
        self.cfgTurnEnabled = tk.BooleanVar(value = self.cfg.natConfig.turnEnabled)
        self.cfgTurnServer = tk.StringVar(value = self.cfg.natConfig.turnServer)
        self.cfgTurnConnType = tk.IntVar(value = self.cfg.natConfig.turnConnType)
        self.cfgTurnUser = tk.StringVar(value = self.cfg.natConfig.turnUserName)
        self.cfgTurnPasswd = tk.StringVar(value = self.cfg.natConfig.turnPassword)

        # Build the tab page
        frm = ttk.Frame(self.frm)
        frm.columnconfigure(0, weight=1)
        frm.columnconfigure(1, weight=2)
        row = 0
        ttk.Label(frm, text='SIP STUN Usage:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Radiobutton(frm, text='Default', value=pj.PJSUA_STUN_USE_DEFAULT, variable=self.cfgSipUseStun).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Disable', value=pj.PJSUA_STUN_USE_DISABLED, variable=self.cfgSipUseStun).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='Media STUN Usage:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Radiobutton(frm, text='Default', value=pj.PJSUA_STUN_USE_DEFAULT, variable=self.cfgMediaUseStun).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='Disable', value=pj.PJSUA_STUN_USE_DISABLED, variable=self.cfgMediaUseStun).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='ICE:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Checkbutton(frm, text='Enable', variable=self.cfgIceEnabled).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Checkbutton(frm, text='Use aggresive nomination', variable=self.cfgIceAggressive).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Checkbutton(frm, text='Always re-INVITE after negotiation', variable=self.cfgAlwaysUpdate).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Checkbutton(frm, text='Disable host candidates', variable=self.cfgIceNoHostCands).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Label(frm, text='TURN:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Checkbutton(frm, text='Enable', variable=self.cfgTurnEnabled).grid(row=row, column=1, sticky=tk.W, padx=6, pady=2)
        row += 1
        ttk.Label(frm, text='TURN server:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Entry(frm, textvariable=self.cfgTurnServer, width=20).grid(row=row, column=1, sticky=tk.W, padx=6)
        ttk.Label(frm, text='host[:port]').grid(row=row, column=1, sticky=tk.E, pady=6)
        row += 1
        ttk.Label(frm, text='TURN connection:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Radiobutton(frm, text='UDP', value=pj.PJ_TURN_TP_UDP, variable=self.cfgTurnConnType).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Radiobutton(frm, text='TCP', value=pj.PJ_TURN_TP_TCP, variable=self.cfgTurnConnType).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='TURN username:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Entry(frm, textvariable=self.cfgTurnUser, width=16).grid(row=row, column=1, sticky=tk.W, padx=6)
        row += 1
        ttk.Label(frm, text='TURN password:').grid(row=row, column=0, sticky=tk.E, pady=2)
        ttk.Entry(frm, textvariable=self.cfgTurnPasswd, show='*', width=16).grid(row=row, column=1, sticky=tk.W, padx=6)

        self.wTab.add(frm, text='NAT settings')

    def onOk(self):
        # Check basic settings
        errors = "";
        if not self.cfgAccId.get():
            errors += "Account ID is required\n"
        if self.cfgAccId.get():
            if not endpoint.validateSipUri(self.cfgAccId.get()):
                errors += "Invalid SIP ID URI: '%s'\n" % (self.cfgAccId.get())
        if self.cfgRegistrar.get():
            if not endpoint.validateSipUri(self.cfgRegistrar.get()):
                errors += "Invalid SIP registrar URI: '%s'\n" % (self.cfgRegistrar.get())
        if self.cfgProxy.get():
            if not endpoint.validateSipUri(self.cfgProxy.get()):
                errors += "Invalid SIP proxy URI: '%s'\n" % (self.cfgProxy.get())
        if self.cfgTurnEnabled.get():
            if not self.cfgTurnServer.get():
                errors += "TURN server is required\n"
        if errors:
            msgbox.showerror("Error detected:", errors)
            return

        # Basic settings
        self.cfg.priority = self.cfgPriority.get()
        self.cfg.idUri = self.cfgAccId.get()
        self.cfg.regConfig.registrarUri = self.cfgRegistrar.get()
        self.cfg.regConfig.registerOnAdd = self.cfgRegisterOnAdd.get()
        while len(self.cfg.sipConfig.authCreds):
            self.cfg.sipConfig.authCreds.pop()
        if self.cfgUsername.get():
            cred = pj.AuthCredInfo()
            cred.scheme = "digest"
            cred.realm = "*"
            cred.username = self.cfgUsername.get()
            cred.data = self.cfgPassword.get()
            self.cfg.sipConfig.authCreds.append(cred)
        while len(self.cfg.sipConfig.proxies):
            self.cfg.sipConfig.proxies.pop()
        if self.cfgProxy.get():
            self.cfg.sipConfig.proxies.append(self.cfgProxy.get())

        # SIP features
        self.cfg.callConfig.prackUse		= self.cfgPrackUse.get()
        self.cfg.callConfig.timerUse		= self.cfgTimerUse.get()
        self.cfg.callConfig.timerSessExpiresSec	= self.cfgTimerExpires.get()
        self.cfg.presConfig.publishEnabled	= self.cfgPublish.get()
        self.cfg.mwiConfig.enabled		= self.cfgMwiEnabled.get()
        self.cfg.natConfig.contactRewriteUse	= 1 if self.cfgEnableContactRewrite.get() else 0
        self.cfg.natConfig.viaRewriteUse 	= 1 if self.cfgEnableViaRewrite.get() else 0
        self.cfg.natConfig.sdpNatRewriteUse	= 1 if self.cfgEnableSdpRewrite.get() else 0
        self.cfg.natConfig.sipOutboundUse	= 1 if self.cfgEnableSipOutbound.get() else 0
        self.cfg.natConfig.udpKaIntervalSec	= self.cfgKaInterval.get()

        # Media
        self.cfg.mediaConfig.transportConfig.port	= self.cfgMedPort.get()
        self.cfg.mediaConfig.transportConfig.portRange	= self.cfgMedPortRange.get()
        self.cfg.mediaConfig.lockCodecEnabled		= self.cfgMedLockCodec.get()
        self.cfg.mediaConfig.srtpUse			= self.cfgMedSrtp.get()
        self.cfg.mediaConfig.srtpSecureSignaling	= self.cfgMedSrtpSecure.get()
        self.cfg.mediaConfig.ipv6Use			= pj.PJSUA_IPV6_ENABLED if self.cfgMedIpv6.get() else pj.PJSUA_IPV6_DISABLED

        # NAT
        self.cfg.natConfig.sipStunUse		= self.cfgSipUseStun.get()
        self.cfg.natConfig.mediaStunUse		= self.cfgMediaUseStun.get()
        self.cfg.natConfig.iceEnabled		= self.cfgIceEnabled.get()
        self.cfg.natConfig.iceAggressiveNomination = self.cfgIceAggressive .get()
        self.cfg.natConfig.iceAlwaysUpdate	= self.cfgAlwaysUpdate.get()
        self.cfg.natConfig.iceMaxHostCands	= 0 if self.cfgIceNoHostCands.get() else -1
        self.cfg.natConfig.turnEnabled		= self.cfgTurnEnabled.get()
        self.cfg.natConfig.turnServer		= self.cfgTurnServer.get()
        self.cfg.natConfig.turnConnType		= self.cfgTurnConnType.get()
        self.cfg.natConfig.turnUserName		= self.cfgTurnUser.get()
        self.cfg.natConfig.turnPasswordType	= 0
        self.cfg.natConfig.turnPassword		= self.cfgTurnPasswd.get()

        self.isOk = True
        self.destroy()

    def onCancel(self):
        self.destroy()


if __name__ == '__main__':
    application.main()



#
# Sample and simple Python script to make and receive calls, and do
# presence and instant messaging/IM using PJSUA-API binding for Python.
#
# Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
#
import py_pjsua
import sys
import thread

#
# Configurations
#
THIS_FILE = "pjsua_app.py"
C_QUIT = 0
C_LOG_LEVEL = 4

# STUN config.
# Set C_STUN_HOST to the address:port of the STUN server to enable STUN
#
C_STUN_HOST = ""
#C_STUN_HOST = "192.168.0.2"
#C_STUN_HOST = "stun.iptel.org:3478"

# SIP port
C_SIP_PORT = 5060


# Globals
#
g_ua_cfg = None
g_acc_id = py_pjsua.PJSUA_INVALID_ID
g_current_call = py_pjsua.PJSUA_INVALID_ID
g_wav_files = []
g_wav_id = 0
g_wav_port = 0
g_rec_file = ""
g_rec_id = 0
g_rec_port = 0

# Utility: display PJ error and exit
#
def err_exit(title, rc):
    py_pjsua.perror(THIS_FILE, title, rc)
    py_pjsua.destroy()
    exit(1)


# Logging function (also callback, called by pjsua-lib)
#
def log_cb(level, str, len):
    if level <= C_LOG_LEVEL:
        print str,

def write_log(level, str):
    log_cb(level, str + "\n", 0)


# Utility to get call info
#
def call_name(call_id):
	ci = py_pjsua.call_get_info(call_id)
	return "[Call " + `call_id` + " " + ci.remote_info + "]"

# Callback when call state has changed.
#
def on_call_state(call_id, e):	
	global g_current_call
	ci = py_pjsua.call_get_info(call_id)
	write_log(3, call_name(call_id) + " state = " + `ci.state_text`)
	if ci.state == py_pjsua.PJSIP_INV_STATE_DISCONNECTED:
		g_current_call = py_pjsua.PJSUA_INVALID_ID

# Callback for incoming call
#
def on_incoming_call(acc_id, call_id, rdata):
	global g_current_call
	
	if g_current_call != py_pjsua.PJSUA_INVALID_ID:
		# There's call in progress - answer Busy
		py_pjsua.call_answer(call_id, 486, None, None)
		return
	
	g_current_call = call_id
	ci = py_pjsua.call_get_info(call_id)
	write_log(3, "*** Incoming call: " + call_name(call_id) + "***")
	write_log(3, "*** Press a to answer or h to hangup  ***")
	
	
	
# Callback when media state has changed (e.g. established or terminated)
#
def on_call_media_state(call_id):
	ci = py_pjsua.call_get_info(call_id)
	if ci.media_status == py_pjsua.PJSUA_CALL_MEDIA_ACTIVE:
		py_pjsua.conf_connect(ci.conf_slot, 0)
		py_pjsua.conf_connect(0, ci.conf_slot)
		write_log(3, call_name(call_id) + ": media is active")
	else:
		write_log(3, call_name(call_id) + ": media is inactive")


# Callback when account registration state has changed
#
def on_reg_state(acc_id):
	acc_info = py_pjsua.acc_get_info(acc_id)
	if acc_info.has_registration != 0:
		cmd = "registration"
	else:
		cmd = "unregistration"
	if acc_info.status != 0 and acc_info.status != 200:
		write_log(3, "Account " + cmd + " failed: rc=" + `acc_info.status` + " " + acc_info.status_text)
	else:
		write_log(3, "Account " + cmd + " success")


# Callback when buddy's presence state has changed
#
def on_buddy_state(buddy_id):
	write_log(3, "On Buddy state called")
	buddy_info = py_pjsua.buddy_get_info(buddy_id)
	if buddy_info.status != 0 and buddy_info.status != 200:
		write_log(3, "Status of " + `buddy_info.uri` + " is " + `buddy_info.status_text`)
	else:
		write_log(3, "Status : " + `buddy_info.status`)

# Callback on incoming pager (MESSAGE)
#		
def on_pager(call_id, strfrom, strto, contact, mime_type, text):
	write_log(3, "MESSAGE from " + `strfrom` + " : " + `text`)


# Callback on the delivery status of outgoing pager (MESSAGE)
#	
def on_pager_status(call_id, strto, body, user_data, status, reason):
	write_log(3, "MESSAGE to " + `strto` + " status " + `status` + " reason " + `reason`)


# Received typing indication
#
def on_typing(call_id, strfrom, to, contact, is_typing):
	str_t = ""
	if is_typing:
		str_t = "is typing.."
	else:
		str_t = "has stopped typing"
	write_log(3, "IM indication: " + strfrom + " " + str_t)

# Received the status of previous call transfer request
#
def on_call_transfer_status(call_id,status_code,status_text,final,p_cont):
	strfinal = ""
	if final == 1:
		strfinal = "[final]"
	
	write_log(3, "Call " + `call_id` + ": transfer status= " + `status_code` + " " + status_text+ " " + strfinal)
	      
	if status_code/100 == 2:
		write_log(3, "Call " + `call_id` + " : call transferred successfully, disconnecting call")
		status = py_pjsua.call_hangup(call_id, 410, None, None)
		p_cont = 0

# Callback on incoming call transfer request
#		
def on_call_transfer_request(call_id, dst, code):
	write_log(3, "Call transfer request from " + `call_id` + " to " + dst + " with code " + `code`)

#
# Initialize pjsua.
#
def app_init():
	global g_acc_id, g_ua_cfg

	# Create pjsua before anything else
	status = py_pjsua.create()
	if status != 0:
		err_exit("pjsua create() error", status)

	# Create and initialize logging config
	log_cfg = py_pjsua.logging_config_default()
	log_cfg.level = C_LOG_LEVEL
	log_cfg.cb = log_cb

	# Create and initialize pjsua config
	# Note: for this Python module, thread_cnt must be 0 since Python
	#       doesn't like to be called from alien thread (pjsua's thread
	#       in this case)	    
	ua_cfg = py_pjsua.config_default()
	ua_cfg.thread_cnt = 0
	ua_cfg.user_agent = "PJSUA/Python 0.1"
	ua_cfg.cb.on_incoming_call = on_incoming_call
	ua_cfg.cb.on_call_media_state = on_call_media_state
	ua_cfg.cb.on_reg_state = on_reg_state
	ua_cfg.cb.on_call_state = on_call_state
	ua_cfg.cb.on_buddy_state = on_buddy_state
	ua_cfg.cb.on_pager = on_pager
	ua_cfg.cb.on_pager_status = on_pager_status
	ua_cfg.cb.on_typing = on_typing
	ua_cfg.cb.on_call_transfer_status = on_call_transfer_status
	ua_cfg.cb.on_call_transfer_request = on_call_transfer_request

	# Configure STUN setting
	if C_STUN_HOST != "":
	    ua_cfg.stun_host = C_STUN_HOST;

	# Create and initialize media config
	med_cfg = py_pjsua.media_config_default()
	med_cfg.ec_tail_len = 0

	#
	# Initialize pjsua!!
	#
	status = py_pjsua.init(ua_cfg, log_cfg, med_cfg)
	if status != 0:
		err_exit("pjsua init() error", status)

	# Configure UDP transport config
	transport_cfg = py_pjsua.transport_config_default()
	transport_cfg.port = C_SIP_PORT

	# Create UDP transport
	status, transport_id = \
	    py_pjsua.transport_create(py_pjsua.PJSIP_TRANSPORT_UDP, transport_cfg)
	if status != 0:
		err_exit("Error creating UDP transport", status)
		

	# Create initial default account
	status, acc_id = py_pjsua.acc_add_local(transport_id, 1)
	if status != 0:
		err_exit("Error creating account", status)

	g_acc_id = acc_id
	g_ua_cfg = ua_cfg

# Add SIP account interractively
#
def add_account():
	global g_acc_id

	acc_domain = ""
	acc_username = ""
	acc_passwd =""
	confirm = ""
	
	# Input account configs
	print "Your SIP domain (e.g. myprovider.com): ",
	acc_domain = sys.stdin.readline()
	if acc_domain == "\n": 
		return
	acc_domain = acc_domain.replace("\n", "")

	print "Your username (e.g. alice): ",
	acc_username = sys.stdin.readline()
	if acc_username == "\n":
		return
	acc_username = acc_username.replace("\n", "")

	print "Your password (e.g. secret): ",
	acc_passwd = sys.stdin.readline()
	if acc_passwd == "\n":
		return
	acc_passwd = acc_passwd.replace("\n", "")

	# Configure account configuration
	acc_cfg = py_pjsua.acc_config_default()
	acc_cfg.id = "sip:" + acc_username + "@" + acc_domain
	acc_cfg.reg_uri = "sip:" + acc_domain

	cred_info = py_pjsua.Pjsip_Cred_Info()
	cred_info.realm = "*"
	cred_info.scheme = "digest"
	cred_info.username = acc_username
	cred_info.data_type = 0
	cred_info.data = acc_passwd

	acc_cfg.cred_info.append(1)
	acc_cfg.cred_info[0] = cred_info

	# Add new SIP account
	status, acc_id = py_pjsua.acc_add(acc_cfg, 1)
	if status != 0:
		py_pjsua.perror(THIS_FILE, "Error adding SIP account", status)
	else:
		g_acc_id = acc_id
		write_log(3, "Account " + acc_cfg.id + " added")

def add_player():
	global g_wav_files
	global g_wav_id
	global g_wav_port
	
	file_name = ""
	status = -1
	wav_id = 0
	
	print "Enter the path of the file player(e.g. /tmp/audio.wav): ",
	file_name = sys.stdin.readline()
	if file_name == "\n": 
		return
	file_name = file_name.replace("\n", "")
	status, wav_id = py_pjsua.player_create(file_name, 0)
	if status != 0:
		py_pjsua.perror(THIS_FILE, "Error adding file player ", status)
	else:
		g_wav_files.append(file_name)
		if g_wav_id == 0:
			g_wav_id = wav_id
			g_wav_port = py_pjsua.player_get_conf_port(wav_id)
		write_log(3, "File player " + file_name + " added")
		
def add_recorder():
	global g_rec_file
	global g_rec_id
	global g_rec_port
	
	file_name = ""
	status = -1
	rec_id = 0
	
	print "Enter the path of the file recorder(e.g. /tmp/audio.wav): ",
	file_name = sys.stdin.readline()
	if file_name == "\n": 
		return
	file_name = file_name.replace("\n", "")
	status, rec_id = py_pjsua.recorder_create(file_name, 0, None, 0, 0)
	if status != 0:
		py_pjsua.perror(THIS_FILE, "Error adding file recorder ", status)
	else:
		g_rec_file = file_name
		g_rec_id = rec_id
		g_rec_port = py_pjsua.recorder_get_conf_port(rec_id)
		write_log(3, "File recorder " + file_name + " added")

def conf_list():
	ports = None
	print "Conference ports : "
	ports = py_pjsua.enum_conf_ports()

	for port in ports:
		info = None
		info = py_pjsua.conf_get_port_info(port)
		txlist = ""
		for listener in info.listeners:
			txlist = txlist + "#" + `listener` + " "
		
		print "Port #" + `info.slot_id` + "[" + `(info.clock_rate/1000)` + "KHz/" + `(info.samples_per_frame * 1000 / info.clock_rate)` + "ms] " + info.name + " transmitting to: " + txlist
		
def connect_port():
	src_port = 0
	dst_port = 0
	
	print "Connect src port # (empty to cancel): "
	src_port = sys.stdin.readline()
	if src_port == "\n": 
		return
	src_port = src_port.replace("\n", "")
	src_port = int(src_port)
	print "To dst port # (empty to cancel): "
	dst_port = sys.stdin.readline()
	if dst_port == "\n": 
		return
	dst_port = dst_port.replace("\n", "")
	dst_port = int(dst_port)
	status = py_pjsua.conf_connect(src_port, dst_port)
	if status != 0:
		py_pjsua.perror(THIS_FILE, "Error connecting port ", status)
	else:		
		write_log(3, "Port connected from " + `src_port` + " to " + `dst_port`)
		
def disconnect_port():
	src_port = 0
	dst_port = 0
	
	print "Disconnect src port # (empty to cancel): "
	src_port = sys.stdin.readline()
	if src_port == "\n": 
		return
	src_port = src_port.replace("\n", "")
	src_port = int(src_port)
	print "From dst port # (empty to cancel): "
	dst_port = sys.stdin.readline()
	if dst_port == "\n": 
		return
	dst_port = dst_port.replace("\n", "")
	dst_port = int(dst_port)
	status = py_pjsua.conf_disconnect(src_port, dst_port)
	if status != 0:
		py_pjsua.perror(THIS_FILE, "Error disconnecting port ", status)
	else:		
		write_log(3, "Port disconnected " + `src_port` + " from " + `dst_port`)

def dump_call_quality():
	global g_current_call
	
	buf = ""
	if g_current_call != -1:
		buf = py_pjsua.call_dump(g_current_call, 1, 1024, "  ")
		write_log(3, "\n" + buf)
	else:
		write_log(3, "No current call")

def xfer_call():
	global g_current_call
	
	if g_current_call == -1:
		
		write_log(3, "No current call")

	else:
		call = g_current_call		
		ci = py_pjsua.call_get_info(g_current_call)
		print "Transferring current call ["+ `g_current_call` + "] " + ci.remote_info
		print "Enter sip url : "
		url = sys.stdin.readline()
		if url == "\n": 
			return
		url = url.replace("\n", "")
		if call != g_current_call:
			print "Call has been disconnected"
			return
		msg_data = py_pjsua.msg_data_init()
		status = py_pjsua.call_xfer(g_current_call, url, msg_data);
		if status != 0:
			py_pjsua.perror(THIS_FILE, "Error transferring call ", status)
		else:		
			write_log(3, "Call transferred to " + url)
		
def xfer_call_replaces():
	if g_current_call == -1:
		write_log(3, "No current call")
	else:
		call = g_current_call
		
		ids = py_pjsua.enum_calls()
		if len(ids) <= 1:
			print "There are no other calls"
			return		

		ci = py_pjsua.call_get_info(g_current_call)
		print "Transfer call [" + `g_current_call` + "] " + ci.remote_info + " to one of the following:"
		for i in range(0, len(ids)):		    
			if ids[i] == call:
				continue
			call_info = py_pjsua.call_get_info(ids[i])
			print `ids[i]` + "  " +  call_info.remote_info + "  [" + call_info.state_text + "]"		

		print "Enter call number to be replaced : "
		buf = sys.stdin.readline()
		buf = buf.replace("\n","")
		if buf == "":
			return
		dst_call = int(buf)
		
		if call != g_current_call:
			print "Call has been disconnected"
			return		
		
		if dst_call == call:
			print "Destination call number must not be the same as the call being transferred"
			return
		
		if dst_call >= py_pjsua.PJSUA_MAX_CALLS:
			print "Invalid destination call number"
			return
	
		if py_pjsua.call_is_active(dst_call) == 0:
			print "Invalid destination call number"
			return		

		py_pjsua.call_xfer_replaces(call, dst_call, 0, None)
		
#
# Worker thread function.
# Python doesn't like it when it's called from an alien thread
# (pjsua's worker thread, in this case), so for Python we must
# disable worker thread in pjsua and poll pjsua from Python instead.
#
def worker_thread_main(arg):
	global C_QUIT
	thread_desc = 0;
	status = py_pjsua.thread_register("python worker", thread_desc)
	if status != 0:
		py_pjsua.perror(THIS_FILE, "Error registering thread", status)
	else:
		while C_QUIT == 0:
			py_pjsua.handle_events(50)
		print "Worker thread quitting.."
		C_QUIT = 2


# Start pjsua
#
def app_start():
	# Done with initialization, start pjsua!!
	#
	status = py_pjsua.start()
	if status != 0:
		err_exit("Error starting pjsua!", status)

	# Start worker thread
	thr = thread.start_new(worker_thread_main, (0,))
    
	print "PJSUA Started!!"


# Print account and buddy list
def print_acc_buddy_list():
	global g_acc_id
	
	acc_ids = py_pjsua.enum_accs()
	print "Account list:"
	for acc_id in acc_ids:
		acc_info = py_pjsua.acc_get_info(acc_id)
		if acc_info.has_registration == 0:
			acc_status = acc_info.status_text
		else:
			acc_status = `acc_info.status` + "/" + acc_info.status_text + " (expires=" + `acc_info.expires` + ")"

		if acc_id == g_acc_id:
			print " *",
		else:
			print "  ",

		print "[" + `acc_id` + "] " + acc_info.acc_uri + ": " + acc_status
		print "       Presence status: ",
		if acc_info.online_status != 0:
			print "Online"
		else:
			print "Invisible"

	if py_pjsua.get_buddy_count() > 0:
		print ""
		print "Buddy list:"
		buddy_ids = py_pjsua.enum_buddies()
		for buddy_id in buddy_ids:
			bi = py_pjsua.buddy_get_info(buddy_id)
			print "   [" + `buddy_id` + "] " + bi.status_text + " " + bi.uri
	
		
# Print application menu
#
def print_menu():
	print ""
	print ">>>"
	print_acc_buddy_list()
	print """
+============================================================================+
|         Call Commands :      |  Buddy, IM & Presence:   |    Account:      |
|                              |                          |                  |
|  m  Make call                | +b  Add buddy            | +a  Add account  |
|  a  Answer current call      | -b  Delete buddy         | -a  Delete accnt |
|  h  Hangup current call      |                          |                  |
|  H  Hold call                |  i  Send instant message | rr  register     |
|  v  re-inVite (release Hold) |  s  Subscribe presence   | ru  Unregister   |
|  #  Send DTMF string         |  u  Unsubscribe presence |                  |
| dq  Dump curr. call quality  |  t  ToGgle Online status |                  |
|                              +--------------------------+------------------+
|  x  Xfer call                |     Media Commands:      |    Status:       |
|  X  Xfer with Replaces       |                          |                  |
|                              | cl  List ports           |  d  Dump status  |
|                              | cc  Connect port         | dd  Dump detail  |
|                              | cd  Disconnect port      |                  |
|                              | +p  Add file player      |                  |
|------------------------------+ +r  Add file recorder    |                  |
|  q  Quit application         |                          |                  |
+============================================================================+"""
	print "You have " + `py_pjsua.call_get_count()` + " active call(s)"
	print ">>>", 

# Menu
#
def app_menu():
	global g_acc_id
	global g_current_call

	quit = 0
	while quit == 0:
		print_menu()
		choice = sys.stdin.readline()

		if choice[0] == "q":
			quit = 1

		elif choice[0] == "i":
			# Sending IM	
			print "Send IM to SIP URL: ",
			url = sys.stdin.readline()
			if url == "\n":
				continue

			# Send typing indication
			py_pjsua.im_typing(g_acc_id, url, 1, None) 

			print "The content: ",
			message = sys.stdin.readline()
			if message == "\n":
				py_pjsua.im_typing(g_acc_id, url, 0, None) 		
				continue

			# Send the IM!
			py_pjsua.im_send(g_acc_id, url, None, message, None, 0)

		elif choice[0] == "m":
			# Make call 
			print "Using account ", g_acc_id
			print "Make call to SIP URL: ",
			url = sys.stdin.readline()
			url = url.replace("\n", "")
			if url == "":
				continue

			# Initiate the call!
			status, call_id = py_pjsua.call_make_call(g_acc_id, url, 0, 0, None)
            
			if status != 0:
				py_pjsua.perror(THIS_FILE, "Error making call", status)
			else:
				g_current_call = call_id

		elif choice[0] == "+" and choice[1] == "b":
			# Add new buddy
			bc = py_pjsua.Buddy_Config()
			print "Buddy URL: ",
			bc.uri = sys.stdin.readline()
			if bc.uri == "\n":
				continue
            
			bc.uri = bc.uri.replace("\n", "")
			bc.subscribe = 1
			status, buddy_id = py_pjsua.buddy_add(bc)
			if status != 0:
				py_pjsua.perror(THIS_FILE, "Error adding buddy", status)
		elif choice[0] == "-" and choice[1] == "b":
			print "Enter buddy ID to delete : "
			buf = sys.stdin.readline()
			buf = buf.replace("\n","")
			if buf == "":
				continue
			i = int(buf)
			if py_pjsua.buddy_is_valid(i) == 0:
				print "Invalid buddy id " + `i`
			else:
				py_pjsua.buddy_del(i)
				print "Buddy " + `i` + " deleted"		
		elif choice[0] == "+" and choice[1] == "a":
			# Add account
			add_account()
		elif choice[0] == "-" and choice[1] == "a":
			print "Enter account ID to delete : "
			buf = sys.stdin.readline()
			buf = buf.replace("\n","")
			if buf == "":
				continue
			i = int(buf)

			if py_pjsua.acc_is_valid(i) == 0:
				print "Invalid account id " + `i`
			else:
				py_pjsua.acc_del(i)
				print "Account " + `i` + " deleted"
	    
		elif choice[0] == "+" and choice[1] == "p":
			add_player()
		elif choice[0] == "+" and choice[1] == "r":
			add_recorder()
		elif choice[0] == "c" and choice[1] == "l":
			conf_list()
		elif choice[0] == "c" and choice[1] == "c":
			connect_port()
		elif choice[0] == "c" and choice[1] == "d":
			disconnect_port()
		elif choice[0] == "d" and choice[1] == "q":
			dump_call_quality()
		elif choice[0] == "x":
			xfer_call()
		elif choice[0] == "X":
			xfer_call_replaces()
		elif choice[0] == "h":
			if g_current_call != py_pjsua.PJSUA_INVALID_ID:
				py_pjsua.call_hangup(g_current_call, 603, None, None)
			else:
				print "No current call"
		elif choice[0] == "H":
			if g_current_call != py_pjsua.PJSUA_INVALID_ID:
				py_pjsua.call_set_hold(g_current_call, None)
		
			else:
				print "No current call"
		elif choice[0] == "v":
			if g_current_call != py_pjsua.PJSUA_INVALID_ID:
		
				py_pjsua.call_reinvite(g_current_call, 1, None);

			else:
				print "No current call"
		elif choice[0] == "#":
			if g_current_call == py_pjsua.PJSUA_INVALID_ID:		
				print "No current call"
			elif py_pjsua.call_has_media(g_current_call) == 0:
				print "Media is not established yet!"
			else:
				call = g_current_call
				print "DTMF strings to send (0-9*#A-B)"
				buf = sys.stdin.readline()
				buf = buf.replace("\n", "")
				if buf == "":
					continue
				if call != g_current_call:
					print "Call has been disconnected"
					continue
				status = py_pjsua.call_dial_dtmf(g_current_call, buf)
				if status != 0:
					py_pjsua.perror(THIS_FILE, "Unable to send DTMF", status);
				else:
					print "DTMF digits enqueued for transmission"
		elif choice[0] == "s":
			print "Subscribe presence of (buddy id) : "
			buf = sys.stdin.readline()
			buf = buf.replace("\n","")
			if buf == "":
				continue
			i = int(buf)
			py_pjsua.buddy_subscribe_pres(i, 1)
		elif choice[0] == "u":
			print "Unsubscribe presence of (buddy id) : "
			buf = sys.stdin.readline()
			buf = buf.replace("\n","")
			if buf == "":
				continue
			i = int(buf)
			py_pjsua.buddy_subscribe_pres(i, 0)
		elif choice[0] == "t":
			acc_info = py_pjsua.acc_get_info(g_acc_id)
			if acc_info.online_status == 0:
				acc_info.online_status = 1
			else:
				acc_info.online_status = 0
			py_pjsua.acc_set_online_status(g_acc_id, acc_info.online_status)
			st = ""
			if acc_info.online_status == 0:
				st = "offline"
			else:
				st = "online"
			print "Setting " + acc_info.acc_uri + " online status to " + st
		elif choice[0] == "r":
			if choice[1] == "r":	    
				py_pjsua.acc_set_registration(g_acc_id, 1)
			elif choice[1] == "u":
				py_pjsua.acc_set_registration(g_acc_id, 0)
		elif choice[0] == "d":
			py_pjsua.dump(choice[1] == "d")
		elif choice[0] == "a":
			if g_current_call != py_pjsua.PJSUA_INVALID_ID:				
				
				py_pjsua.call_answer(g_current_call, 200, None, None)
			else:
				print "No current call"


#
# main
#
app_init()
app_start()
app_menu()

#
# Done, quitting..
#
print "PJSUA shutting down.."
C_QUIT = 1
# Give the worker thread chance to quit itself
while C_QUIT != 2:
    py_pjsua.handle_events(50)

print "PJSUA destroying.."
py_pjsua.destroy()




import inc_sip as sip
import inc_sdp as sdp

sdp = \
"""
v=0
o=- 0 0 IN IP4 127.0.0.1
s=pjmedia
c=IN IP4 127.0.0.1
t=0 0
m=audio 4000 RTP/AVP 0 101
a=rtpmap:0 PCMU/8000
a=sendrecv
a=rtpmap:101 telephone-event/8000
a=fmtp:101 0-15
"""

sendto_cfg = sip.SendtoCfg( "simple call", "--null-audio --auto-answer 200", sdp, 200)




import inc_sip as sip
import inc_sdp as sdp

# In this test we simulate broken server, where it always sends
# stale=true with all 401 responses. We should expect pjsip to
# retry the authentication until PJSIP_MAX_STALE_COUNT is
# exceeded. When pjsip retries the authentication, it should
# use the new nonce from server


pjsua = "--null-audio --id=sip:CLIENT --registrar sip:127.0.0.1:$PORT " + \
	"--realm=python --user=username --password=password"

req1 = sip.RecvfromTransaction("Initial request", 401,
				include=["REGISTER sip"], 
				exclude=["Authorization"],
				resp_hdr=["WWW-Authenticate: Digest realm=\"python\", nonce=\"1\""]
			  	)

req2 = sip.RecvfromTransaction("First retry", 401,
				include=["REGISTER sip", "Authorization", "nonce=\"1\""], 
				exclude=["Authorization:[\\s\\S]+Authorization:"],
				resp_hdr=["WWW-Authenticate: Digest realm=\"python\", nonce=\"2\", stale=true"]
			  	)

req3 = sip.RecvfromTransaction("Second retry retry", 401,
				include=["REGISTER sip", "Authorization", "nonce=\"2\""], 
				exclude=["Authorization:[\\s\\S]+Authorization:"],
				resp_hdr=["WWW-Authenticate: Digest realm=\"python\", nonce=\"3\", stale=true"]
				)

req4 = sip.RecvfromTransaction("Third retry", 401,
				include=["REGISTER sip", "Authorization", "nonce=\"3\""], 
				exclude=["Authorization:[\\s\\S]+Authorization:"],
				resp_hdr=["WWW-Authenticate: Digest realm=\"python\", nonce=\"4\", stale=true"],
				expect="PJSIP_EAUTHSTALECOUNT"
			  	)

recvfrom_cfg = sip.RecvfromCfg("Failed registration retry (server rejects with stale=true) ",
			       pjsua, [req1, req2, req3, req4])



import sys
import inc_sip as sip
import inc_const as const
import inc_util as util
import re
from inc_cfg import *

# Read configuration
cfg_file = util.load_module_from_file("cfg_file", ARGS[1])

# Test body function
def test_func(t):
	pjsua = t.process[0]
	# Create dialog
	dlg = sip.Dialog("127.0.0.1", pjsua.inst_param.sip_port, 
			  tcp=cfg_file.sendto_cfg.use_tcp)
	#dlg = sip.Dialog("127.0.0.1", 5060, tcp=cfg_file.sendto_cfg.use_tcp)
	cfg = cfg_file.sendto_cfg
	
	if len(cfg.complete_msg) != 0:
		req = dlg.update_fields(cfg.complete_msg)
	else:
		req = dlg.create_invite(cfg.sdp, cfg.extra_headers, cfg.body)
	resp = dlg.send_request_wait(req, 10)
	if resp=="":
		raise TestError("Timed-out waiting for response")
	# Check response code
	code = int(sip.get_code(resp))
	if code != cfg.resp_code:
		dlg.hangup(code)
		raise TestError("Expecting code " + str(cfg.resp_code) + 
				" got " + str(code))
	# Check for patterns that must exist
	for p in cfg.resp_include:
		if re.search(p, resp, re.M | re.I)==None:
			dlg.hangup(code)
			raise TestError("Pattern " + p + " not found")
	# Check for patterns that must not exist
	for p in cfg.resp_exclude:
		if re.search(p, resp, re.M | re.I)!=None:
			dlg.hangup(code)
			raise TestError("Excluded pattern " + p + " found")
	pjsua.sync_stdout()
	dlg.hangup(code)
	pjsua.sync_stdout()

# Here where it all comes together
test = TestParam(cfg_file.sendto_cfg.name, 
		 [cfg_file.sendto_cfg.inst_param], 
		 test_func)





#
from inc_cfg import *

# Trickle ICE full vs regular, should be okay (as there are host candidates)
test_param = TestParam(
		"Callee=Regular ICE, caller=Trickle ICE (full)",
		[
			InstanceParam("callee", "--null-audio --max-calls=1 --use-ice --ice-trickle=0 --stun-srv stun.pjsip.org"),
			InstanceParam("caller", "--null-audio --max-calls=1 --use-ice --ice-trickle=2 --stun-srv stun.pjsip.org")
		]
		)



#
# Just about the simple pjsua command line parameter, which should
# never fail in any circumstances
from inc_cfg import *

test_param = TestParam(
		"Basic run", 
		[
			InstanceParam("pjsua", "--null-audio --rtp-port 0")
		]
		)




#
# pjsua Python GUI Demo
#
# Copyright (C)2013 Teluu Inc. (http://www.teluu.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
#
import sys
if sys.version_info[0] >= 3: # Python 3
    import tkinter as tk
    from tkinter import ttk
    from tkinter import messagebox as msgbox
else:
    import Tkinter as tk
    import tkMessageBox as msgbox
    import ttk

import pjsua2 as pj
import application


class Endpoint(pj.Endpoint):
    """
    This is high level Python object inherited from pj.Endpoint
    """
    instance = None
    def __init__(self):
        pj.Endpoint.__init__(self)
        Endpoint.instance = self


def validateUri(uri):
    return Endpoint.instance.utilVerifyUri(uri) == pj.PJ_SUCCESS

def validateSipUri(uri):
    return Endpoint.instance.utilVerifySipUri(uri) == pj.PJ_SUCCESS


if __name__ == '__main__':
    application.main()



import inc_sip as sip
import inc_sdp as sdp

# Too long key should be rejected
sdp = \
"""
v=0
o=- 0 0 IN IP4 127.0.0.1
s=-
c=IN IP4 127.0.0.1
t=0 0
m=audio 5000 RTP/SAVP 0
a=crypto:1 aes_cm_128_hmac_sha1_80 inline:WnD7c1ksDGs+dIefCEo8omPg4uO8DYIinNGL5yxQWnD7c1ksDGs+dIefCEo8omPg4uO8DYIinNGL5yxQ
"""

pjsua_args = "--null-audio --auto-answer 200 --use-srtp 2 --srtp-secure 0"
extra_headers = ""
include = []
exclude = []

sendto_cfg = sip.SendtoCfg("SRTP receive too long key", pjsua_args, sdp, 406,
			   extra_headers=extra_headers,
			   resp_inc=include, resp_exc=exclude) 




import inc_sip as sip
import inc_sdp as sdp

sdp = \
"""
v=0
o=- 0 0 IN IP4 127.0.0.1
s=pjmedia
c=IN IP4 127.0.0.1
t=0 0
m=audio 4000 RTP/AVP 0 101
a=rtpmap:0 PCMU/8000
a=sendrecv
a=rtpmap:101 telephone-event/8000
a=fmtp:101 0-15
"""

# RFC 4028 Section 9:
#   If the incoming request contains a Supported header field with a
#   value 'timer' but does not contain a Session-Expires header, it means
#   that the UAS is indicating support for timers but is not requesting
#   one.  The UAS may request a session timer in the 2XX response by
#   including a Session-Expires header field.  The value MUST NOT be set
#   to a duration lower than the value in the Min-SE header field in the
#   request, if it is present.
   
pjsua_args = "--null-audio --auto-answer 200 --use-timer 2 --timer-min-se 90 --timer-se 1800"
extra_headers = "Supported: timer\n"
include = ["Session-Expires: .*;refresher=.*"]
exclude = []
sendto_cfg = sip.SendtoCfg("Session Timer initiated by UAS", pjsua_args, sdp, 200, 
			   extra_headers=extra_headers,
			   resp_inc=include, resp_exc=exclude) 
			   




import sys
import os

REMOVE_THESE = ["-I/usr/include", "-I/usr/include/", "-L/usr/lib", "-L/usr/lib/"]

class Pkg:
	def __init__(self, pkg_name):
		self.name = pkg_name
		self.priority = 0
		self.vars = {}

	def parse(self, pkg_config_path):
		f = None
		for pkg_path in pkg_config_path.split(':'):
			if pkg_path[-1] != '/':
				pkg_path += '/'
			fname = pkg_path + self.name + '.pc'
			try:
				f = open(fname, "r")
				break
			except:
				continue
		if not f:
		  	#sys.stderr.write("pkgconfig.py: unable to find %s.pc in %s\n" % (self.name, pkg_config_path))
			return False

		for line in f.readlines():
			line = line.strip()
			if not line:
				continue
			if line[0]=='#':
				continue
			pos1 = line.find('=')
			pos2 = line.find(':')
			if pos1 > 0 and (pos1 < pos2 or pos2 < 0):
				pos = pos1
			elif pos2 > 0 and (pos2 < pos1 or pos1 < 0):
				pos = pos2
			else:
				continue
			name = line[:pos].lower()
			value = line[pos+1:]
			self.vars[name] = value
		f.close()

		for name in self.vars.keys():
			value = self.vars[name]
			while True:
				pos1 = value.find("${")
				if pos1 < 0:
					break
				pos2 = value.find("}")
				if pos2 < 0:
					break
				value = value.replace(value[pos1:pos2+1], self.vars[value[pos1+2:pos2]])
			self.vars[name] = value
		return True

	def requires(self):
		if not 'requires' in self.vars:
			return []
		deps = []
		req_list = self.vars['requires']
		for req_item in req_list.split(','):
			req_item = req_item.strip()
			for i in range(len(req_item)):
				if "=<>".find(req_item[i]) >= 0:
					deps.append(req_item[:i].strip())
					break
		return deps

	def libs(self):
		if not 'libs' in self.vars:
			return []
		return self.vars['libs'].split(' ')

	def cflags(self):
		if not 'cflags' in self.vars:
			return []
		return self.vars['cflags'].split(' ')


def calculate_pkg_priority(pkg, pkg_dict, loop_cnt):
	if loop_cnt > 10:
		sys.stderr.write("Circular dependency with pkg %s\n" % (pkg))
		return 0
	reqs = pkg.requires()
	prio = 1
	for req in reqs:
		if not req in pkg_dict:
			continue
		req_pkg = pkg_dict[req]
		prio += calculate_pkg_priority(req_pkg, pkg_dict, loop_cnt+1)
	return prio

if __name__ == "__main__":
	pkg_names = []
	pkg_dict = {}
	commands = []
	exist_check = False

	for i in range(1,len(sys.argv)):
		if sys.argv[i][0] == '-':
			cmd = sys.argv[i]
			commands.append(cmd)
			if cmd=='--exists':
				exist_check = True
			elif cmd=="--help":
				print "This is not very helpful, is it"
				sys.exit(0)
			elif cmd=="--version":
				print "0.1"
				sys.exit(0)
		else:
			pkg_names.append(sys.argv[i])
			
	# Fix search path
	PKG_CONFIG_PATH = os.getenv("PKG_CONFIG_PATH", "").strip()
	if not PKG_CONFIG_PATH:
		PKG_CONFIG_PATH="/usr/local/lib/pkgconfig:/usr/lib/pkgconfig"
	PKG_CONFIG_PATH = PKG_CONFIG_PATH.replace(";", ":")

	# Parse files
	for pkg_name in pkg_names:
		pkg = Pkg(pkg_name)
		if not pkg.parse(PKG_CONFIG_PATH):
			sys.exit(1)
		pkg_dict[pkg_name] = pkg

	if exist_check:
		sys.exit(0)

	# Calculate priority based on dependency
	for pkg_name in pkg_dict.keys():
		pkg = pkg_dict[pkg_name]
		pkg.priority = calculate_pkg_priority(pkg, pkg_dict, 1)
	
	# Sort package based on dependency
	pkg_names = sorted(pkg_names, key=lambda pkg_name: pkg_dict[pkg_name].priority, reverse=True)

	# Get the options
	opts = []
	for cmd in commands:
		if cmd=='--libs':
			for pkg_name in pkg_names:
				libs = pkg_dict[pkg_name].libs()
				for lib in libs:
					opts.append(lib)
					if lib[:2]=="-l":
						break
			for pkg_name in pkg_names:
				opts += pkg_dict[pkg_name].libs()
		elif cmd=='--cflags':
			for pkg_name in pkg_names:
				opts += pkg_dict[pkg_name].cflags()
		elif cmd[0]=='-':
			sys.stderr.write("pkgconfig.py: I don't know how to handle " + sys.argv[i] + "\n")
	
	filtered_opts = []
	for opt in opts:
		opt = opt.strip()
		if not opt:
			continue
		if REMOVE_THESE.count(opt) != 0:
			continue
		if opt != '-framework' and opt != '--framework' and filtered_opts.count(opt) != 0:
			if len(filtered_opts) and (filtered_opts[-1] == '-framework' or filtered_opts[-1] == '--framework'):
				filtered_opts.pop()
			continue
		filtered_opts.append(opt)

	print ' '.join(filtered_opts)





## Automatic test module for SIPp.
##
## This module will need a test driver for each SIPp scenario:
## - For simple scenario, i.e: make/receive call (including auth), this
##   test module can auto-generate a default test driver, i.e: make call
##   or apply auto answer. Just name the SIPp scenario using "uas" or
##   "uac" prefix accordingly.
## - Custom test driver can be defined in a python script file containing
##   a list of the PJSUA instances and another list for PJSUA expects/
##   commands. The custom test driver file must use the same filename as
##   the SIPp XML scenario. See samples of SIPp scenario + its driver
##   in tests/pjsua/scripts-sipp/ folder for detail.
##
##   Here are defined macros that can be used in the custom driver:
##   - $SIPP_PORT            : SIPp binding port
##   - $SIPP_URI            : SIPp SIP URI
##   - $PJSUA_PORT[N]            : binding port of PJSUA instance #N
##   - $PJSUA_URI[N]            : SIP URI of PJSUA instance #N
import ctypes
import time
import sys
import os
import re
import subprocess
from inc_cfg import *
import inc_const
import inc_util as util

# flags that test is running in Unix
G_INUNIX = False
if sys.platform.lower().find("win32")!=-1 or sys.platform.lower().find("microsoft")!=-1:
    G_INUNIX = False
else:
    G_INUNIX = True

# /dev/null handle, for redirecting output when SIPP is not in background mode
FDEVNULL = None

# SIPp executable path and param
#SIPP_PATH = '"C:\\devs\\bin\\Sipp_3.2\\sipp.exe"'
SIPP_PATH = 'sipp'
SIPP_PORT    = 50070
SIPP_PARAM = "-m 1 -i 127.0.0.1 -p " + str(SIPP_PORT)
SIPP_TIMEOUT = 60
# On BG mode, SIPp doesn't require special terminal
# On non-BG mode, on win, it needs env var: "TERMINFO=c:\cygwin\usr\share\terminfo"
# TODO: on unix with BG mode, waitpid() always fails, need to be fixed
SIPP_BG_MODE = False
#SIPP_BG_MODE = not G_INUNIX

# Will be updated based on the test driver file (a .py file whose the same name as SIPp XML file)
PJSUA_INST_PARAM = []
PJSUA_EXPECTS = []

# Default PJSUA param if test driver is not available:
# - no-tcp as SIPp is on UDP only
# - id, username, and realm: to allow PJSUA sending re-INVITE with auth after receiving 401/407 response
PJSUA_DEF_PARAM = "--null-audio --max-calls=1 --no-tcp --id=sip:a@localhost --username=a --realm=*"

# Get SIPp scenario (XML file)
SIPP_SCEN_XML  = ""
if ARGS[1].endswith('.xml'):
    SIPP_SCEN_XML  = ARGS[1]
else:
    exit(-99)


# Functions for resolving macros in the test driver
def resolve_pjsua_port(mo):
    return str(PJSUA_INST_PARAM[int(mo.group(1))].sip_port)

def resolve_pjsua_uri(mo):
    return PJSUA_INST_PARAM[int(mo.group(1))].uri[1:-1]

def resolve_driver_macros(st):
    st = re.sub("\$SIPP_PORT", str(SIPP_PORT), st)
    st = re.sub("\$SIPP_URI", "sip:sipp@127.0.0.1:"+str(SIPP_PORT), st)
    st = re.sub("\$PJSUA_PORT\[(\d+)\]", resolve_pjsua_port, st)
    st = re.sub("\$PJSUA_URI\[(\d+)\]", resolve_pjsua_uri, st)
    return st


# Init test driver
if os.access(SIPP_SCEN_XML[:-4]+".py", os.R_OK):
    # Load test driver file (the corresponding .py file), if any
    cfg_file = util.load_module_from_file("cfg_file", SIPP_SCEN_XML[:-4]+".py")
    for ua_idx, ua_param in enumerate(cfg_file.PJSUA):
        ua_param = resolve_driver_macros(ua_param)
        PJSUA_INST_PARAM.append(InstanceParam("pjsua"+str(ua_idx), ua_param))
    if DEFAULT_TELNET and hasattr(cfg_file, 'PJSUA_CLI_EXPECTS'):
        PJSUA_EXPECTS = cfg_file.PJSUA_CLI_EXPECTS
    else:
        PJSUA_EXPECTS = cfg_file.PJSUA_EXPECTS
else:
    # Generate default test driver
    if os.path.basename(SIPP_SCEN_XML)[0:3] == "uas":
        # auto make call when SIPp is as UAS
        ua_param = PJSUA_DEF_PARAM + " sip:127.0.0.1:" + str(SIPP_PORT)
    else:
        # auto answer when SIPp is as UAC
        ua_param = PJSUA_DEF_PARAM + " --auto-answer=200" 
    PJSUA_INST_PARAM.append(InstanceParam("pjsua", ua_param))


# Start SIPp process, returning PID
def start_sipp():
    global SIPP_BG_MODE
    sipp_proc = None

    sipp_param = SIPP_PARAM + " -sf " + SIPP_SCEN_XML
    if SIPP_BG_MODE:
        sipp_param = sipp_param + " -bg"
    if SIPP_TIMEOUT:
        sipp_param = sipp_param + " -timeout "+str(SIPP_TIMEOUT)+"s -timeout_error" + " -deadcall_wait "+str(SIPP_TIMEOUT)+"s"

    # add target param
    sipp_param = sipp_param + " 127.0.0.1:" + str(PJSUA_INST_PARAM[0].sip_port)

    # run SIPp
    fullcmd = os.path.normpath(SIPP_PATH) + " " + sipp_param
    print("Running SIPP: " + fullcmd)
    if SIPP_BG_MODE:
        sipp_proc = subprocess.Popen(fullcmd, bufsize=0, stdin=subprocess.PIPE, stdout=subprocess.PIPE, shell=G_INUNIX, universal_newlines=False)
    else:
        # redirect output to NULL
        global FDEVNULL
        #FDEVNULL  = open(os.devnull, 'w')
        FDEVNULL  = open("logs/sipp_output.tmp", 'w')
        sipp_proc = subprocess.Popen(fullcmd, shell=G_INUNIX, stdout=FDEVNULL, stderr=FDEVNULL)

    if not SIPP_BG_MODE:
        if sipp_proc == None or sipp_proc.poll():
            return None
        return sipp_proc

    else:
        # get SIPp child process PID
        pid = 0
        r = re.compile("PID=\[(\d+)\]", re.I)

        while True:
            line = sipp_proc.stdout.readline()
            pid_r = r.search(line)
            if pid_r:
                pid = int(pid_r.group(1))
                break
            if not sipp_proc.poll():
                break

        if pid != 0:
            # Win specific: get process handle from PID, as on win32, os.waitpid() takes process handle instead of pid
            if (sys.platform == "win32"):
                SYNCHRONIZE = 0x00100000
                PROCESS_QUERY_INFORMATION = 0x0400
                hnd = ctypes.windll.kernel32.OpenProcess(SYNCHRONIZE | PROCESS_QUERY_INFORMATION, False, pid)
                pid = hnd

        return pid


# Wait SIPp process to exit, returning SIPp exit code
def wait_sipp(sipp):
    if not SIPP_BG_MODE:
        global FDEVNULL
        sipp.wait()
        FDEVNULL.close()
        return sipp.returncode

    else:
        print("Waiting SIPp (PID=" + str(sipp) + ") to exit..")
        wait_cnt = 0
        while True:
            try:
                wait_cnt = wait_cnt + 1
                [pid_, ret_code] = os.waitpid(sipp, 0)
                if sipp == pid_:
                    #print "SIPP returned ", ret_code
                    ret_code = ret_code >> 8

                    # Win specific: Close process handle
                    if (sys.platform == "win32"):
                        ctypes.windll.kernel32.CloseHandle(sipp)
                    
                    return ret_code
            except os.error:
                if wait_cnt <= 5:
                    print("Retry ("+str(wait_cnt)+") waiting SIPp..")
                else:
                    return -99


# Execute PJSUA flow
def exec_pjsua_expects(t, sipp):
    # Get all PJSUA instances
    ua = []
    for ua_idx in range(len(PJSUA_INST_PARAM)):
        ua.append(t.process[ua_idx])

    ua_err_st = ""
    while len(PJSUA_EXPECTS):
        expect = PJSUA_EXPECTS.pop(0)
        ua_idx = expect[0]
        expect_st = expect[1]
        send_cmd = resolve_driver_macros(expect[2])
        timeout = expect[3] if len(expect)>=4 else 0
        # Handle exception in pjsua flow, to avoid zombie SIPp process
        try:
            if expect_st != "":
                if timeout > 0:
                    ua[ua_idx].expect(expect_st, raise_on_error = True, timeout = timeout)
                else:
                    ua[ua_idx].expect(expect_st, raise_on_error = True)
            if send_cmd != "":
                ua[ua_idx].send(send_cmd)
        except TestError as e:
            ua_err_st = e.desc
            break;
        except:
            ua_err_st = "Unknown error"
            break;

    # Need to poll here for handling these cases:
    # - If there is no PJSUA EXPECT scenario, we must keep polling the stdout,
    #   otherwise PJSUA process may stuck (due to stdout pipe buffer full?).
    # - last PJSUA_EXPECT contains a pjsua command that needs time to
    #   finish, for example "v" (re-INVITE), the SIPp XML scenario may expect
    #   that re-INVITE transaction to be completed and without stdout poll
    #   PJSUA process may stuck.
    # Ideally the poll should be done contiunously until SIPp process is
    # terminated.
    # Update: now pjsua stdout is polled continuously by a dedicated thread,
    #         so the poll is no longer needed
    #for ua_idx in range(len(ua)):
    #        ua[ua_idx].expect(inc_const.STDOUT_REFRESH, raise_on_error = False)

    return ua_err_st


def sipp_err_to_str(err_code):
    if err_code == 0:
        return "All calls were successful"
    elif err_code == 1:
        return "At least one call failed"
    elif err_code == 97:
        return "exit on internal command. Calls may have been processed"
    elif err_code == 99:
        return "Normal exit without calls processed"
    elif err_code == -1:
        return "Fatal error (timeout)"
    elif err_code == -2:
        return "Fatal error binding a socket"
    else:
        return "Unknown error"


# Test body function
def TEST_FUNC(t):

    sipp_ret_code = 0
    ua_err_st = ""

    sipp = start_sipp()
    if not sipp:
        raise TestError("Failed starting SIPp")

    ua_err_st = exec_pjsua_expects(t, sipp)

    sipp_ret_code = wait_sipp(sipp)

    if ua_err_st != "":
        raise TestError(ua_err_st)

    if sipp_ret_code:
        rc = ctypes.c_byte(sipp_ret_code).value
        raise TestError("SIPp returned error " + str(rc) + ": " + sipp_err_to_str(rc))


# Here where it all comes together
test = TestParam(SIPP_SCEN_XML[:-4],
                 PJSUA_INST_PARAM,
                 TEST_FUNC)



#
# importsym.py: Import C symbol decls (structs, enums, etc) and write them
#               to another file
#
# Copyright (C)2013 Teluu Inc. (http://www.teluu.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import pycparser
from pycparser import c_generator
import sys
import os

def which(program):
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    if sys.platform == 'win32' and not program.endswith(".exe"):
        program += ".exe"

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None

#
PJ_ROOT_PATH = "../../../"

# CPP is needed by pycparser.
CPP_PATH = which("cpp")
if not CPP_PATH:
    print 'Error: need to have cpp in PATH'
    sys.exit(1)

# Hardcoded!
# Note for win32:
# - use Windows Command Prompt app, not msys/mingw
# - temporarily comment "#include <pj/compat/socket.h>" in pj/sock.h (line ~29)
if sys.platform == 'win32':
    PYCPARSER_DIR="D:/work/tool/pycparser-master"
elif sys.platform == "linux2":
    PYCPARSER_DIR="/home/bennylp/Desktop/opt/src/pycparser-master"
else:
    PYCPARSER_DIR="/Library/Python/2.7/site-packages/pycparser"

if not os.path.exists(PYCPARSER_DIR + '/utils/fake_libc_include'):
    print "Error: couldn't find pycparser utils in '%s'" % PYCPARSER_DIR
    sys.exit(1)

# Heading, to be placed before the source files
C_HEADING_SECTION = """
#define PJ_AUTOCONF        1
#define jmp_buf            int
#define __attribute__(x)
"""

# CPP (C preprocessor) settings
CPP_CFLAGS   = [
    '-I' + PYCPARSER_DIR + '/utils/fake_libc_include',
    "-I" + PJ_ROOT_PATH + "pjlib/include",
    "-I" + PJ_ROOT_PATH + "pjlib-util/include",
    "-I" + PJ_ROOT_PATH + "pjnath/include",
    "-I" + PJ_ROOT_PATH + "pjmedia/include",
    "-I" + PJ_ROOT_PATH + "pjsip/include"
    ]


class SymbolVisitor(pycparser.c_ast.NodeVisitor):
    def __init__(self, names):
        self.nodeDict = {}
        for name in names:
            self.nodeDict[name] = None

    def _add(self, node):
        if self.nodeDict.has_key(node.name):
            self.nodeDict[node.name] = node

    def visit_Struct(self, node):
        self._add(node)

    def visit_Enum(self, node):
        self._add(node)

    def visit_Typename(self, node):
        self._add(node)

    def visit_Typedef(self, node):
        self._add(node)


TEMP_FILE="tmpsrc.h"

class SymbolImporter:
    """
    Import C selected declarations from C source file and move it
    to another file.

    Parameters:
     - listfile    Path of file containing list of C source file
                    and identifier names to be imported. The format
                    of the listfile is:

                    filename        name1  name2  name3

                    for example:

                    pj/sock_qos.h    pj_qos_type  pj_qos_flag
                    pj/types.h    pj_status_t  PJ_SUCCESS
    """
    def __init__(self):
        pass

    def process(self, listfile, outfile):

        # Read listfile
        f = open(listfile)
        lines = f.readlines()
        f.close()

        # Process each line in list file, while generating the
        # temporary C file to be processed by pycparser
        f = open(TEMP_FILE, "w")
        f.write(C_HEADING_SECTION)
        names = []
        fcnt = 0
        for line in lines:
            spec = line.split()
            if len(spec) < 2:
                continue
            fcnt += 1
            f.write("#include <%s>\n" % spec[0])
            names.extend(spec[1:])
        f.close()
        print 'Parsing %d symbols from %d files..' % (len(names), fcnt)

        try:
            # Parse the temporary C file
            ast = pycparser.parse_file(TEMP_FILE, use_cpp=True, cpp_path=CPP_PATH, cpp_args=CPP_CFLAGS)
            os.remove(TEMP_FILE)

            # Filter the declarations that we wanted
            print 'Filtering..'
            visitor = SymbolVisitor(names)
            visitor.visit(ast)

            # Print symbol declarations to outfile
            print 'Writing declarations..'
            f = open(outfile, 'w')
            f.write("// This file is autogenerated by importsym script, do not modify!\n\n")
            gen = pycparser.c_generator.CGenerator()
            for name in names:
                node = visitor.nodeDict[name]
                if not node:
                    print "  warning: declaration for '%s' is not found **" % k
                else:
                    print "  writing '%s'.." % name
                    output = gen.visit(node) + ";\n\n"
                    f.write(output)
            f.close()
            print "Done."

        except Exception as e:
            print e

if __name__ == "__main__":
    print "Importing symbols: 'symbols.lst' --> 'symbols.i'"
    si = SymbolImporter()
    si.process("symbols.lst", "symbols.i")
    try:
        os.remove("lextab.py")
    except OSError:
        pass
    try:
        os.remove("yacctab.py")
    except OSError:
        pass



import inc_sip as sip
import inc_sdp as sdp

sdp = \
"""
v=0
o=- 0 0 IN IP4 127.0.0.1
s=tester
c=IN IP4 127.0.0.1
t=0 0
m=audio 4000 RTP/AVP 0 101
a=rtpmap:0 PCMU/8000
a=sendrecv
a=rtpmap:101 telephone-event/8000
a=fmtp:101 0-15
a=crypto:1 AES_CM_128_HMAC_SHA1_80 inline:WnD7c1ksDGs+dIefCEo8omPg4uO8DYIinNGL5yxQ
a=crypto:1 AES_CM_128_HMAC_SHA1_32 inline:t0r0/apkukU7JjjfR0mY8GEimBq4OiPEm9eKSFOx
"""

args = "--null-audio --auto-answer 200 --max-calls 1 --use-srtp 1 --srtp-secure 0"
include = []
exclude = []

sendto_cfg = sip.SendtoCfg( "caller has used invalid crypto tag, callee must not accept the call", 
			    pjsua_args=args, sdp=sdp, resp_code=406, 
			    resp_inc=include, resp_exc=exclude)



# Useful constants


##########################
# MENU OUTPUT
#


##########################
# EVENTS
#

# Text to expect when there is incoming call
EVENT_INCOMING_CALL = "Press .* answer"


##########################
# CALL STATES
#

# Call state is CALLING
STATE_CALLING = "state.*CALLING"
# Call state is EARLY
STATE_EARLY = "state.*EARLY"
# Call state is CONFIRMED
STATE_CONFIRMED = "state.*CONFIRMED"
# Call state is DISCONNECTED
STATE_DISCONNECTED = "Call .* DISCONNECTED"

# Media call is put on-hold
MEDIA_HOLD = "Call [0-9]+ media [0-9]+ .*, status is .* hold"
# Media call is active
MEDIA_ACTIVE = "Call [0-9]+ media [0-9]+ .*, status is Active"
#MEDIA_ACTIVE = "Media for call [0-9]+ is active"
# RX_DTMF
RX_DTMF = "Incoming DTMF on call [0-9]+: "

##########################
# MEDIA
#

# Connecting/disconnecting ports
MEDIA_CONN_PORT_SUCCESS = "Port \d+ \(.+\) transmitting to port"
MEDIA_DISCONN_PORT_SUCCESS = "Port \d+ \(.+\) stop transmitting to port"

# Filename to play / record
MEDIA_PLAY_FILE = "--play-file\s+(\S+)"
MEDIA_REC_FILE = "--rec-file\s+(\S+)"

##########################
# MISC
#

# The command prompt
PROMPT = ">>>"
# When pjsua has been destroyed
DESTROYED = "PJSUA destroyed"
# Assertion failure
ASSERT = "Assertion failed"
# Stdout refresh text
STDOUT_REFRESH = "XXSTDOUT_REFRESHXX"





#
from inc_cfg import *

# Different number of ICE components
test_param = TestParam(
		"Callee=use ICE, caller=use ICE",
		[
			InstanceParam("callee", "--null-audio --use-ice --max-calls=1"),
			InstanceParam("caller", "--null-audio --use-ice --max-calls=1 --ice-no-rtcp")
		]
		)



#
# pjsua Python GUI Demo
#
# Copyright (C)2013 Teluu Inc. (http://www.teluu.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
#
import sys
if sys.version_info[0] >= 3: # Python 3
    import tkinter as tk
    from tkinter import ttk
else:
    import Tkinter as tk
    import ttk

import buddy
import call
import chatgui as gui
import endpoint as ep
import pjsua2 as pj
import re

SipUriRegex = re.compile('(sip|sips):([^:;>\@]*)@?([^:;>]*):?([^:;>]*)')
ConfIdx = 1
write=sys.stdout.write

# Simple SIP uri parser, input URI must have been validated
def ParseSipUri(sip_uri_str):
    m = SipUriRegex.search(sip_uri_str)
    if not m:
        assert(0)
        return None

    scheme = m.group(1)
    user = m.group(2)
    host = m.group(3)
    port = m.group(4)
    if host == '':
        host = user
        user = ''

    return SipUri(scheme.lower(), user, host.lower(), port)

class SipUri:
    def __init__(self, scheme, user, host, port):
        self.scheme = scheme
        self.user = user
        self.host = host
        self.port = port

    def __cmp__(self, sip_uri):
        if self.scheme == sip_uri.scheme and self.user == sip_uri.user and self.host == sip_uri.host:
            # don't check port, at least for now
            return 0
        return -1

    def __str__(self):
        s = self.scheme + ':'
        if self.user: s += self.user + '@'
        s += self.host
        if self.port: s+= ':' + self.port
        return s

class Chat(gui.ChatObserver):
    def __init__(self, app, acc, uri, call_inst=None):
        self._app = app
        self._acc = acc
        self.title = ''

        global ConfIdx
        self.confIdx = ConfIdx
        ConfIdx += 1

        # each participant call/buddy instances are stored in call list
        # and buddy list with same index as in particpant list
        self._participantList = []	# list of SipUri
        self._callList = []		# list of Call
        self._buddyList = []		# list of Buddy

        self._gui = gui.ChatFrame(self)
        self.addParticipant(uri, call_inst)

    def _updateGui(self):
        if self.isPrivate():
            self.title = str(self._participantList[0])
        else:
            self.title = 'Conference #%d (%d participants)' % (self.confIdx, len(self._participantList))
        self._gui.title(self.title)
        self._app.updateWindowMenu()

    def _getCallFromUriStr(self, uri_str, op = ''):
        uri = ParseSipUri(uri_str)
        if uri not in self._participantList:
            write("=== "+ op +" cannot find participant with URI '" + uri_str + "'\r\n")
            return None
        idx = self._participantList.index(uri)
        if idx < len(self._callList):
            return self._callList[idx]
        return None

    def _getActiveMediaIdx(self, thecall):
        ci = thecall.getInfo()
        for mi in ci.media:
            if mi.type == pj.PJMEDIA_TYPE_AUDIO and \
              (mi.status != pj.PJSUA_CALL_MEDIA_NONE and \
               mi.status != pj.PJSUA_CALL_MEDIA_ERROR):
                return mi.index
        return -1

    def _getAudioMediaFromUriStr(self, uri_str):
        c = self._getCallFromUriStr(uri_str)
        if not c: return None

        idx = self._getActiveMediaIdx(c)
        if idx < 0: return None

        m = c.getMedia(idx)
        am = pj.AudioMedia.typecastFromMedia(m)
        return am

    def _sendTypingIndication(self, is_typing, sender_uri_str=''):
        sender_uri = ParseSipUri(sender_uri_str) if sender_uri_str else None
        type_ind_param = pj.SendTypingIndicationParam()
        type_ind_param.isTyping = is_typing
        for idx, p in enumerate(self._participantList):
            # don't echo back to the original sender
            if sender_uri and p == sender_uri:
                continue

            # send via call, if any, or buddy
            target = None
            if self._callList[idx] and self._callList[idx].connected:
                target = self._callList[idx]
            else:
                target = self._buddyList[idx]
            assert(target)

            try:
                target.sendTypingIndication(type_ind_param)
            except:
                pass

    def _sendInstantMessage(self, msg, sender_uri_str=''):
        sender_uri = ParseSipUri(sender_uri_str) if sender_uri_str else None
        send_im_param = pj.SendInstantMessageParam()
        send_im_param.content = str(msg)
        for idx, p in enumerate(self._participantList):
            # don't echo back to the original sender
            if sender_uri and p == sender_uri:
                continue

            # send via call, if any, or buddy
            target = None
            if self._callList[idx] and self._callList[idx].connected:
                target = self._callList[idx]
            else:
                target = self._buddyList[idx]
            assert(target)

            try:
                target.sendInstantMessage(send_im_param)
            except:
                # error will be handled via Account::onInstantMessageStatus()
                pass

    def isPrivate(self):
        return len(self._participantList) <= 1

    def isUriParticipant(self, uri):
        return uri in self._participantList

    def registerCall(self, uri_str, call_inst):
        uri = ParseSipUri(uri_str)
        try:
            idx = self._participantList.index(uri)
            bud = self._buddyList[idx]
            self._callList[idx] = call_inst
            call_inst.chat = self
            call_inst.peerUri = bud.cfg.uri
        except:
            assert(0) # idx must be found!

    def showWindow(self, show_text_chat = False):
        self._gui.bringToFront()
        if show_text_chat:
            self._gui.textShowHide(True)

    def addParticipant(self, uri, call_inst=None):
        # avoid duplication
        if self.isUriParticipant(uri): return

        uri_str = str(uri)

        # find buddy, create one if not found (e.g: for IM/typing ind),
        # it is a temporary one and not really registered to acc
        bud = None
        try:
            bud = self._acc.findBuddy2(uri_str)
        except:
            bud = buddy.Buddy(None)
            bud_cfg = pj.BuddyConfig()
            bud_cfg.uri = uri_str
            bud_cfg.subscribe = False
            bud.create(self._acc, bud_cfg)
            bud.cfg = bud_cfg
            bud.account = self._acc

        # update URI from buddy URI
        uri = ParseSipUri(bud.cfg.uri)

        # add it
        self._participantList.append(uri)
        self._callList.append(call_inst)
        self._buddyList.append(bud)
        self._gui.addParticipant(str(uri))
        self._updateGui()

    def kickParticipant(self, uri):
        if (not uri) or (uri not in self._participantList):
            assert(0)
            return

        idx = self._participantList.index(uri)
        del self._participantList[idx]
        del self._callList[idx]
        del self._buddyList[idx]
        self._gui.delParticipant(str(uri))

        if self._participantList:
            self._updateGui()
        else:
            self.onCloseWindow()

    def addMessage(self, from_uri_str, msg):
        if from_uri_str:
            # print message on GUI
            msg = from_uri_str + ': ' + msg
            self._gui.textAddMessage(msg)
            # now relay to all participants
            self._sendInstantMessage(msg, from_uri_str)
        else:
            self._gui.textAddMessage(msg, False)

    def setTypingIndication(self, from_uri_str, is_typing):
        # notify GUI
        self._gui.textSetTypingIndication(from_uri_str, is_typing)
        # now relay to all participants
        self._sendTypingIndication(is_typing, from_uri_str)

    def startCall(self):
        self._gui.enableAudio()
        call_param = pj.CallOpParam()
        call_param.opt.audioCount = 1
        call_param.opt.videoCount = 0
        fails = []
        for idx, p in enumerate(self._participantList):
            # just skip if call is instantiated
            if self._callList[idx]:
                continue

            uri_str = str(p)
            c = call.Call(self._acc, uri_str, self)
            self._callList[idx] = c
            self._gui.audioUpdateState(uri_str, gui.AudioState.INITIALIZING)

            try:
                c.makeCall(uri_str, call_param)
            except:
                self._callList[idx] = None
                self._gui.audioUpdateState(uri_str, gui.AudioState.FAILED)
                fails.append(p)

        for p in fails:
            # kick participants with call failure, but spare the last (avoid zombie chat)
            if not self.isPrivate():
                self.kickParticipant(p)

    def stopCall(self):
        for idx, p in enumerate(self._participantList):
            self._gui.audioUpdateState(str(p), gui.AudioState.DISCONNECTED)
            c = self._callList[idx]
            if c:
                c.hangup(pj.CallOpParam())

    def updateCallState(self, thecall, info = None):
        # info is optional here, just to avoid calling getInfo() twice (in the caller and here)
        if not info: info = thecall.getInfo()

        if info.state < pj.PJSIP_INV_STATE_CONFIRMED:
            self._gui.audioUpdateState(thecall.peerUri, gui.AudioState.INITIALIZING)
        elif info.state == pj.PJSIP_INV_STATE_CONFIRMED:
            self._gui.audioUpdateState(thecall.peerUri, gui.AudioState.CONNECTED)
            if not self.isPrivate():
                # inform peer about conference participants
                conf_welcome_str  = '\n---\n'
                conf_welcome_str += 'Welcome to the conference, participants:\n'
                conf_welcome_str += '%s (host)\n' % (self._acc.cfg.idUri)
                for p in self._participantList:
                    conf_welcome_str += '%s\n' % (str(p))
                conf_welcome_str += '---\n'
                send_im_param = pj.SendInstantMessageParam()
                send_im_param.content = conf_welcome_str
                try:
                    thecall.sendInstantMessage(send_im_param)
                except:
                    pass

                # inform others, including self
                msg = "[Conf manager] %s has joined" % (thecall.peerUri)
                self.addMessage(None, msg)
                self._sendInstantMessage(msg, thecall.peerUri)

        elif info.state == pj.PJSIP_INV_STATE_DISCONNECTED:
            if info.lastStatusCode/100 != 2:
                self._gui.audioUpdateState(thecall.peerUri, gui.AudioState.FAILED)
            else:
                self._gui.audioUpdateState(thecall.peerUri, gui.AudioState.DISCONNECTED)

            # reset entry in the callList
            try:
                idx = self._callList.index(thecall)
                if idx >= 0: self._callList[idx] = None
            except:
                pass

            self.addMessage(None, "Call to '%s' disconnected: %s" % (thecall.peerUri, info.lastReason))

            # kick the disconnected participant, but the last (avoid zombie chat)
            if not self.isPrivate():
                self.kickParticipant(ParseSipUri(thecall.peerUri))

                # inform others, including self
                msg = "[Conf manager] %s has left" % (thecall.peerUri)
                self.addMessage(None, msg)
                self._sendInstantMessage(msg, thecall.peerUri)

    def updateCallMediaState(self, thecall, info = None):
        # info is optional here, just to avoid calling getInfo() twice (in the caller and here)
        if not info: info = thecall.getInfo()

        med_idx = self._getActiveMediaIdx(thecall)
        if (med_idx < 0):
            self._gui.audioSetStatsText(thecall.peerUri, 'No active media')
            return

        si = thecall.getStreamInfo(med_idx)
        dir_str = ''
        if si.dir == 0:
            dir_str = 'inactive'
        else:
            if si.dir & pj.PJMEDIA_DIR_ENCODING:
                dir_str += 'send '
            if si.dir & pj.PJMEDIA_DIR_DECODING:
                dir_str += 'receive '
        stats_str  = "Direction   : %s\n" % (dir_str)
        stats_str += "Audio codec : %s (%sHz)" % (si.codecName, si.codecClockRate)
        self._gui.audioSetStatsText(thecall.peerUri, stats_str)
        m = pj.AudioMedia.typecastFromMedia(thecall.getMedia(med_idx))

        # make conference
        for c in self._callList:
            if c == thecall:
                continue
            med_idx = self._getActiveMediaIdx(c)
            if med_idx < 0:
                continue
            mm = pj.AudioMedia.typecastFromMedia(c.getMedia(med_idx))
            m.startTransmit(mm)
            mm.startTransmit(m)


    # ** callbacks from GUI (ChatObserver implementation) **

    # Text
    def onSendMessage(self, msg):
        self._sendInstantMessage(msg)

    def onStartTyping(self):
        self._sendTypingIndication(True)

    def onStopTyping(self):
        self._sendTypingIndication(False)

    # Audio
    def onHangup(self, peer_uri_str):
        c = self._getCallFromUriStr(peer_uri_str, "onHangup()")
        if not c: return
        call_param = pj.CallOpParam()
        c.hangup(call_param)

    def onHold(self, peer_uri_str):
        c = self._getCallFromUriStr(peer_uri_str, "onHold()")
        if not c: return
        call_param = pj.CallOpParam()
        c.setHold(call_param)

    def onUnhold(self, peer_uri_str):
        c = self._getCallFromUriStr(peer_uri_str, "onUnhold()")
        if not c: return

        call_param = pj.CallOpParam()
        call_param.opt.audioCount = 1
        call_param.opt.videoCount = 0
        call_param.opt.flag |= pj.PJSUA_CALL_UNHOLD
        c.reinvite(call_param)

    def onRxMute(self, peer_uri_str, mute):
        am = self._getAudioMediaFromUriStr(peer_uri_str)
        if not am: return
        if mute:
            am.stopTransmit(ep.Endpoint.instance.audDevManager().getPlaybackDevMedia())
            self.addMessage(None, "Muted audio from '%s'" % (peer_uri_str))
        else:
            am.startTransmit(ep.Endpoint.instance.audDevManager().getPlaybackDevMedia())
            self.addMessage(None, "Unmuted audio from '%s'" % (peer_uri_str))

    def onRxVol(self, peer_uri_str, vol_pct):
        am = self._getAudioMediaFromUriStr(peer_uri_str)
        if not am: return
        # pjsua volume range = 0:mute, 1:no adjustment, 2:100% louder
        am.adjustRxLevel(vol_pct/50.0)
        self.addMessage(None, "Adjusted volume level audio from '%s'" % (peer_uri_str))

    def onTxMute(self, peer_uri_str, mute):
        am = self._getAudioMediaFromUriStr(peer_uri_str)
        if not am: return
        if mute:
            ep.Endpoint.instance.audDevManager().getCaptureDevMedia().stopTransmit(am)
            self.addMessage(None, "Muted audio to '%s'" % (peer_uri_str))
        else:
            ep.Endpoint.instance.audDevManager().getCaptureDevMedia().startTransmit(am)
            self.addMessage(None, "Unmuted audio to '%s'" % (peer_uri_str))

    # Chat room
    def onAddParticipant(self):
        buds = []
        dlg = AddParticipantDlg(None, self._app, buds)
        if dlg.doModal():
            for bud in buds:
                uri = ParseSipUri(bud.cfg.uri)
                self.addParticipant(uri)
            if not self.isPrivate():
                self.startCall()

    def onStartAudio(self):
        self.startCall()

    def onStopAudio(self):
        self.stopCall()

    def onCloseWindow(self):
        self.stopCall()
        # will remove entry from list eventually destroy this chat?
        if self in self._acc.chatList: self._acc.chatList.remove(self)
        self._app.updateWindowMenu()
        # destroy GUI
        self._gui.destroy()


class AddParticipantDlg(tk.Toplevel):
    """
    List of buddies
    """
    def __init__(self, parent, app, bud_list):
        tk.Toplevel.__init__(self, parent)
        self.title('Add participants..')
        self.transient(parent)
        self.parent = parent
        self._app = app
        self.buddyList = bud_list

        self.isOk = False

        self.createWidgets()

    def doModal(self):
        if self.parent:
            self.parent.wait_window(self)
        else:
            self.wait_window(self)
        return self.isOk

    def createWidgets(self):
        # buddy list
        list_frame = ttk.Frame(self)
        list_frame.pack(side=tk.TOP, fill=tk.BOTH, expand=1, padx=20, pady=20)
        #scrl = ttk.Scrollbar(self, orient=tk.VERTICAL, command=list_frame.yview)
        #list_frame.config(yscrollcommand=scrl.set)
        #scrl.pack(side=tk.RIGHT, fill=tk.Y)

        # draw buddy list
        self.buddies = []
        for acc in self._app.accList:
            self.buddies.append((0, acc.cfg.idUri))
            for bud in acc.buddyList:
                self.buddies.append((1, bud))

        self.bud_var = []
        for idx,(flag,bud) in enumerate(self.buddies):
            self.bud_var.append(tk.IntVar())
            if flag==0:
                s = ttk.Separator(list_frame, orient=tk.HORIZONTAL)
                s.pack(fill=tk.X)
                l = tk.Label(list_frame, anchor=tk.W, text="Account '%s':" % (bud))
                l.pack(fill=tk.X)
            else:
                c = tk.Checkbutton(list_frame, anchor=tk.W, text=bud.cfg.uri, variable=self.bud_var[idx])
                c.pack(fill=tk.X)
        s = ttk.Separator(list_frame, orient=tk.HORIZONTAL)
        s.pack(fill=tk.X)

        # Ok/cancel buttons
        tail_frame = ttk.Frame(self)
        tail_frame.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=1)

        btnOk = ttk.Button(tail_frame, text='Ok', default=tk.ACTIVE, command=self.onOk)
        btnOk.pack(side=tk.LEFT, padx=20, pady=10)
        btnCancel = ttk.Button(tail_frame, text='Cancel', command=self.onCancel)
        btnCancel.pack(side=tk.RIGHT, padx=20, pady=10)

    def onOk(self):
        self.buddyList[:] = []
        for idx,(flag,bud) in enumerate(self.buddies):
            if not flag: continue
            if self.bud_var[idx].get() and not (bud in self.buddyList):
                self.buddyList.append(bud)

        self.isOk = True
        self.destroy()

    def onCancel(self):
        self.destroy()



#
from inc_cfg import *

# simple test
test_param = TestParam(
		"Resample (large filter) 8 KHZ to 22 KHZ",
		[
			InstanceParam("endpt", "--null-audio --quality 10 --clock-rate 22050 --play-file wavs/input.8.wav --rec-file wavs/tmp.22.wav")
		]
		)



import inc_sip as sip
import inc_sdp as sdp

# Torture message from RFC 4475
# 3.1.1.  Valid Messages
# 3.1.1.3. Valid Use of the % Escaping Mechanism
complete_msg = \
"""INVITE sip:sips%3Auser%40example.com@example.net SIP/2.0
To: sip:%75se%72@example.com
From: <sip:I%20have%20spaces@example.net>;tag=$FROM_TAG
Max-Forwards: 87
i: esc01.239409asdfakjkn23onasd0-3234
CSeq: 234234 INVITE
Via: SIP/2.0/UDP host5.example.net;rport;branch=z9hG4bKkdjuw
C: application/sdp
Contact:
  <sip:cal%6Cer@$LOCAL_IP:$LOCAL_PORT;%6C%72;n%61me=v%61lue%25%34%31>
Content-Length: 150

v=0
o=mhandley 29739 7272939 IN IP4 192.0.2.1
s=-
c=IN IP4 192.0.2.1
t=0 0
m=audio 49217 RTP/AVP 0 12
m=video 3227 RTP/AVP 31
a=rtpmap:31 LPC
"""


sendto_cfg = sip.SendtoCfg( "RFC 4475 3.1.1.3", 
			    "--null-audio --auto-answer 200", 
			    "", 200, complete_msg=complete_msg)




#
from socket import *
import re
import random
import time
import sys
import inc_cfg as cfg
from select import *

# SIP request template
req_templ = \
"""$METHOD $TARGET_URI SIP/2.0\r
Via: SIP/2.0/UDP $LOCAL_IP:$LOCAL_PORT;rport;branch=z9hG4bK$BRANCH\r
Max-Forwards: 70\r
From: <sip:caller@pjsip.org>$FROM_TAG\r
To: <$TARGET_URI>$TO_TAG\r
Contact: <sip:$LOCAL_IP:$LOCAL_PORT;transport=udp>\r
Call-ID: $CALL_ID@pjsip.org\r
CSeq: $CSEQ $METHOD\r
Allow: PRACK, INVITE, ACK, BYE, CANCEL, UPDATE, REFER\r
Supported: replaces, 100rel, norefersub\r
User-Agent: pjsip.org Python tester\r
Content-Length: $CONTENT_LENGTH\r
$SIP_HEADERS"""


def is_request(msg):
	return msg.split(" ", 1)[0] != "SIP/2.0"
	
def is_response(msg):
	return msg.split(" ", 1)[0] == "SIP/2.0"

def get_code(msg):
	if msg=="":
		return 0
	return int(msg.split(" ", 2)[1])

def get_tag(msg, hdr="To"):
	pat = "^" + hdr + ":.*"
	result = re.search(pat, msg, re.M | re.I)
	if result==None:
		return ""
	line = result.group()
	#print "line=", line
	tags = line.split(";tag=")
	if len(tags)>1:
		return tags[1]
	return ""
	#return re.split("[;& ]", s)

def get_header(msg, hname):
	headers = msg.splitlines()
	for hdr in headers:
		hfields = hdr.split(": ", 2)
		if hfields[0]==hname:
			return hfields[1]
	return None

class Dialog:
	sock = None
	dst_addr = ""
	dst_port = 5060
	local_ip = ""
	local_port = 0
	tcp = False
	call_id = str(random.random())
	cseq = 0
	local_tag = ";tag=" + str(random.random())
	rem_tag = ""
	last_resp_code = 0
	inv_branch = ""
	trace_enabled = True
	last_request = ""
	def __init__(self, dst_addr, dst_port=5060, tcp=False, trace=True, local_port=0):
		self.dst_addr = dst_addr
		self.dst_port = dst_port
		self.tcp = tcp
		self.trace_enabled = trace
		if tcp==True:
			self.sock = socket(AF_INET, SOCK_STREAM)
			self.sock.connect(dst_addr, dst_port)
		else:
			self.sock = socket(AF_INET, SOCK_DGRAM)
			self.sock.bind(("127.0.0.1", local_port))
		
		self.local_ip, self.local_port = self.sock.getsockname()
		self.trace("Dialog socket bound to " + self.local_ip + ":" + str(self.local_port))

	def trace(self, txt):
		if self.trace_enabled:
			try:
				print(str(time.strftime("%H:%M:%S ")) + txt)
			except UnicodeEncodeError:
				print((str(time.strftime("%H:%M:%S ")) + txt).encode('utf-8'))

	def update_fields(self, msg):
		if self.tcp:
			transport_param = ";transport=tcp"
		else:
			transport_param = ""
		msg = msg.replace("$TARGET_URI", "sip:"+self.dst_addr+":"+str(self.dst_port) + transport_param)
		msg = msg.replace("$LOCAL_IP", self.local_ip)
		msg = msg.replace("$LOCAL_PORT", str(self.local_port))
		msg = msg.replace("$FROM_TAG", self.local_tag)
		msg = msg.replace("$TO_TAG", self.rem_tag)
		msg = msg.replace("$CALL_ID", self.call_id)
		msg = msg.replace("$CSEQ", str(self.cseq))
		branch=str(random.random())
		msg = msg.replace("$BRANCH", branch)
		return msg

	def create_req(self, method, sdp, branch="", extra_headers="", body=""):
		if branch=="":
			self.cseq = self.cseq + 1
		msg = req_templ
		msg = msg.replace("$METHOD", method)
		msg = msg.replace("$SIP_HEADERS", extra_headers)
		if branch=="":
			branch=str(random.random())
		msg = msg.replace("$BRANCH", branch)
		if sdp!="":
			msg = msg.replace("$CONTENT_LENGTH", str(len(sdp)))
			msg = msg + "Content-Type: application/sdp\r\n"
			msg = msg + "\r\n"
			msg = msg + sdp
		elif body!="":
			msg = msg.replace("$CONTENT_LENGTH", str(len(body)))
			msg = msg + "\r\n"
			msg = msg + body
		else:
			msg = msg.replace("$CONTENT_LENGTH", "0")
		return self.update_fields(msg)

	def create_response(self, request, code, reason, to_tag=""):
		response = "SIP/2.0 " + str(code) + " " + reason + "\r\n"
		lines = request.splitlines()
		for line in lines:
			hdr = line.split(":", 1)[0]
			if hdr in ["Via", "From", "To", "CSeq", "Call-ID"]:
				if hdr=="To" and to_tag!="":
					line = line + ";tag=" + to_tag
				elif hdr=="Via":
					line = line + ";received=127.0.0.1"
				response = response + line + "\r\n"
		return response

	def create_invite(self, sdp, extra_headers="", body=""):
		self.inv_branch = str(random.random())
		return self.create_req("INVITE", sdp, branch=self.inv_branch, extra_headers=extra_headers, body=body)

	def create_ack(self, sdp="", extra_headers=""):
		return self.create_req("ACK", sdp, extra_headers=extra_headers, branch=self.inv_branch)

	def create_bye(self, extra_headers=""):
		return self.create_req("BYE", "", extra_headers)

	def send_msg(self, msg, dst_addr=None):
		if (is_request(msg)):
			self.last_request = msg.split(" ", 1)[0]
		if not dst_addr:
			dst_addr = (self.dst_addr, self.dst_port)
		self.trace("============== TX MSG to " + str(dst_addr) + " ============= \n" + msg)
		self.sock.sendto(msg.encode('utf-8'), 0, dst_addr)

	def wait_msg_from(self, timeout):
		endtime = time.time() + timeout
		msg = ""
		src_addr = None
		while time.time() < endtime:
			readset = select([self.sock], [], [], 1)
			if len(readset[0]) < 1 or not self.sock in readset[0]:
				if len(readset[0]) < 1:
					print("select() timeout (will wait for " + str(int(endtime - time.time())) + "more secs)")
				elif not self.sock in readset[0]:
					print("select() alien socket")
				else:
					print("select other error")
				continue
			try:
				msg, src_addr = self.sock.recvfrom(4096)
				break
			except:
				print("recv() exception: ", sys.exc_info()[0])
				continue
		
		msgstr = msg.decode('utf-8')
		if msgstr=="":
			return "", None
		if self.last_request=="INVITE" and self.rem_tag=="":
			self.rem_tag = get_tag(msgstr, "To")
			self.rem_tag = self.rem_tag.rstrip("\r\n;")
			if self.rem_tag != "":
				self.rem_tag = ";tag=" + self.rem_tag
			self.trace("=== rem_tag:" + self.rem_tag)
		self.trace("=========== RX MSG from " + str(src_addr) +  " ===========\n" + msgstr)
		return (msgstr, src_addr)
	
	def wait_msg(self, timeout):
		return self.wait_msg_from(timeout)[0]
		
	# Send request and wait for final response
	def send_request_wait(self, msg, timeout):
		t1 = 1.0
		endtime = time.time() + timeout
		resp = ""
		code = 0
		for i in range(0,5):
			self.send_msg(msg)
			resp = self.wait_msg(t1)
			if resp!="" and is_response(resp):
				code = get_code(resp)
				break
		last_resp = resp
		while code < 200 and time.time() < endtime:
			resp = self.wait_msg(endtime - time.time())
			if resp != "" and is_response(resp):
				code = get_code(resp)
				last_resp = resp
			elif resp=="":
				break
		return last_resp
	
	def hangup(self, last_code=0):
		self.trace("====== hangup =====")
		if last_code!=0:
			self.last_resp_code = last_code
		if self.last_resp_code>0 and self.last_resp_code<200:
			msg = self.create_req("CANCEL", "", branch=self.inv_branch, extra_headers="")
			self.send_request_wait(msg, 5)
			msg = self.create_ack()
			self.send_msg(msg)
		elif self.last_resp_code>=200 and self.last_resp_code<300:
			msg = self.create_ack()
			self.send_msg(msg)
			msg = self.create_bye()
			self.send_request_wait(msg, 5)
		else:
			msg = self.create_ack()
			self.send_msg(msg)


class SendtoCfg:
	# Test name
	name = ""
	# pjsua InstanceParam
	inst_param = None
	# Complete INVITE message. If this is not empty, then this
	# message will be sent instead and the "sdp" and "extra_headers"
	# settings will be ignored.
	complete_msg = ""
	# Initial SDP
	sdp = ""
	# Extra headers to add to request
	extra_headers = ""
	# Expected code
	resp_code = 0
	# Use TCP?
	use_tcp = False
	# List of RE patterns that must exist in response
	resp_include = []
	# List of RE patterns that must NOT exist in response
	resp_exclude = []
	# Full (non-SDP) body
	body = ""
	# Constructor
	def __init__(self, name, pjsua_args, sdp, resp_code, 
		     resp_inc=[], resp_exc=[], use_tcp=False,
		     extra_headers="", body="", complete_msg="",
		     enable_buffer = False):
		self.complete_msg = complete_msg
		self.sdp = sdp
		self.resp_code = resp_code
		self.resp_include = resp_inc
		self.resp_exclude = resp_exc
		self.use_tcp = use_tcp
		self.extra_headers = extra_headers
		self.body = body
		self.inst_param = cfg.InstanceParam("pjsua", pjsua_args)
		self.inst_param.enable_buffer = enable_buffer 


class RecvfromTransaction:
	# The test title for this transaction
	title = ""
	# Optinal list of pjsua command and optional expect patterns 
	# to be invoked to make pjsua send a request
	# Sample:
	#	(to make call and wait for INVITE to be sent)
	#	cmds = [ ["m"], ["sip:127.0.0.1", "INVITE sip:"]  ]
	cmds = []
	# Check if the CSeq must be greater than last Cseq?
	check_cseq = True
	# List of RE patterns that must exists in incoming request
	include = []
	# List of RE patterns that MUST NOT exist in incoming request
	exclude = []
	# Response code to send
	resp_code = 0
	# Additional list of headers to be sent on the response
	# Note: no need to add CRLF on the header
	resp_hdr = []
	# Message body. This should include the Content-Type header too.
	# Sample:
	#	body = """Content-Type: application/sdp\r\n
	#	\r\n
	#	v=0\r\n
	#	...
	#	"""
	body = None
	# Pattern to be expected on pjsua when receiving the response
	expect = ""
	# Required config
	pj_config = ""
	
	def __init__(self, title, resp_code, check_cseq=True,
			include=[], exclude=[], cmds=[], resp_hdr=[], resp_body=None, expect="", pj_config=""):
		self.title = title
		self.cmds = cmds
		self.include = include
		self.exclude = exclude
		self.resp_code = resp_code
		self.resp_hdr = resp_hdr
		self.body = resp_body
		self.expect = expect
		self.pj_config=pj_config
			

class RecvfromCfg:
	# Test name
	name = ""
	# pjsua InstanceParam
	inst_param = None
	# List of RecvfromTransaction
	transaction = None
	# Use TCP?
	tcp = False
	# Required config
	pj_config = ""

	# Note:
	#  Any "$PORT" string in the pjsua_args will be replaced
	#  by server port
	def __init__(self, name, pjsua_args, transaction, tcp=False, pj_config=""):
		self.name = name
		self.inst_param = cfg.InstanceParam("pjsua", pjsua_args)
		self.transaction = transaction
		self.tcp=tcp
		self.pj_config=pj_config







import inc_sip as sip
import inc_sdp as sdp

pjsua = "--null-audio --id=sip:CLIENT --registrar sip:127.0.0.1:$PORT " + \
	"--realm=* --user=username --password=password " + \
	"--auto-update-nat=0"

# 401 Response, missing realm value
req1 = sip.RecvfromTransaction("Initial request", 401,
				include=["REGISTER sip"], 
				exclude=[],
				resp_hdr=['WWW-Authenticate: Digest']
			  	)

# Client should retry, we giving it another 401 with empty realm
req2 = sip.RecvfromTransaction("REGISTER retry #1 of 2", 407,
				include=["REGISTER sip"], 
				exclude=[],
				resp_hdr=['Proxy-Authenticate: Digest realm=""']
			  	)

# Client should retry
req3 = sip.RecvfromTransaction("REGISTER retry #2 of 2", 200,
				include=[], 
				exclude=[],
				expect="registration success"
			  	)

recvfrom_cfg = sip.RecvfromCfg("Registration with empty realm",
			       pjsua, [req1, req2, req3])



import pjsua2 as pj
import sys
import time
from collections import deque
import struct
import gc
from random import randint
import threading

write=sys.stdout.write

#
# Basic data structure test, to make sure basic struct
# and array operations work
#
def ua_data_test():
    #
    # AuthCredInfo
    #
    write("UA data types test..")
    the_realm = "pjsip.org"
    ci = pj.AuthCredInfo()
    ci.realm = the_realm
    ci.dataType = 20

    ci2 = ci
    assert ci.dataType == 20
    assert ci2.realm == the_realm

    #
    # UaConfig
    # See here how we manipulate std::vector
    #
    uc = pj.UaConfig()
    uc.maxCalls = 10
    uc.userAgent = "Python"
    uc.nameserver = pj.StringVector(["10.0.0.1", "10.0.0.2"])
    uc.nameserver.append("NS1")

    uc2 = uc
    assert uc2.maxCalls == 10
    assert uc2.userAgent == "Python"
    assert len(uc2.nameserver) == 3
    assert uc2.nameserver[0] == "10.0.0.1"
    assert uc2.nameserver[1] == "10.0.0.2"
    assert uc2.nameserver[2] == "NS1"

    write("  Dumping nameservers: " + "\r\n")
    for s in uc2.nameserver:
        write(s  + "\r\n")
    write("\r\n")

    #
    # ByteVector
    # Assign bytes/bytearray/memoryview to a ByteVector
    # Copy the content of a ByteVector to a bytearray/memoryview
    #
    bv = pj.ByteVector()
    bv.assign_from_bytes(bytearray([1, 2, 3]))
    assert bv.size() == 3
    assert bv[0] == 1 and bv[1] == 2 and bv[2] == 3

    ba = memoryview(bytearray(5))
    bv.copy_to_bytearray(ba[1:])
    assert ba == b'\x00\x01\x02\x03\x00'

#
# Exception test
#
def ua_run_test_exception():
    write("Exception test.." + "\r\n")
    ep = pj.Endpoint()
    ep.libCreate()
    got_exception = False
    try:
        ep.natDetectType()
    except pj.Error as e:
        #t, e = sys.exc_info()[:2]
        got_exception = True
        write("  Got exception: status=%u, reason=%s,\n  title=%s,\n  srcFile=%s, srcLine=%d" % \
            (e.status, e.reason, e.title, e.srcFile, e.srcLine) + "\r\n")
        assert e.status == 370050
        assert e.reason.find("PJNATH_ESTUNINSERVER") >= 0
        assert e.title == "pjsua_detect_nat_type()"
    assert got_exception

#
# Custom log writer
#
class MyLogWriter(pj.LogWriter):
    def write(self, entry):
        write("This is Python:" + entry.msg + "\r\n")

class AMP(pj.AudioMediaPort):
    buffers = deque()

    def onFrameRequested(self, frame):
        if len(self.buffers):
            # Get a buffer from the queue
            buffer_out = self.buffers.popleft()
            # Copy the buffer to PJSIP's frame
            frame.type = pj.PJMEDIA_TYPE_AUDIO
            frame.buf.assign_from_bytes(buffer_out)

    def onFrameReceived(self, frame):
        # Get the incoming audio buffer from PJSIP's frame
        buffer_in = bytearray(frame.buf.size())
        frame.buf.copy_to_bytearray(buffer_in)
        # Prepare an output buffer
        buffer_out = bytearray(len(buffer_in))

        # Now you can process these byte arrays directly, e.g. use numpy
        for i in range(len(buffer_in)):
            if (i % 2 == 1):
                # Convert it to signed 16-bit integer
                x = buffer_in[i] << 8 | buffer_in[i-1]
                x = struct.unpack('<h', struct.pack('<H', x))[0]

                # Amplify the signal by 50% and clip it
                x = int(x * 1.5)
                if x > 32767:
                    x = 32767
                elif x < -32768:
                    x = -32768

                # Convert it to unsigned 16-bit integer
                x = struct.unpack('<H', struct.pack('<h', x))[0]

                # Put it in the output buffer in little endian order
                buffer_out[i-1] = (x & 0xff)
                buffer_out[i] = (x & 0xff00) >> 8

        self.buffers.append(buffer_out)

#
# Testing log writer callback
#
def ua_run_log_test():
    write("Logging test.." + "\r\n")
    ep_cfg = pj.EpConfig()

    lw = MyLogWriter()
    ep_cfg.logConfig.writer = lw
    ep_cfg.logConfig.decor = ep_cfg.logConfig.decor & ~(pj.PJ_LOG_HAS_CR | pj.PJ_LOG_HAS_NEWLINE)

    ep = pj.Endpoint()
    ep.libCreate()
    ep.libInit(ep_cfg)
    ep.libDestroy()

#
# Simple create, init, start, and destroy sequence
#
def ua_run_ua_test():
    write("UA test run.." + "\r\n")
    ep_cfg = pj.EpConfig()

    ep = pj.Endpoint()
    ep.libCreate()
    ep.libInit(ep_cfg)
    ep.libStart()

    # Sample custom audio media port
    # Note: threading must be enabled
    amp = AMP()
    fmt = pj.MediaFormatAudio()
    fmt.type = pj.PJMEDIA_TYPE_AUDIO
    fmt.clockRate = 16000
    fmt.channelCount = 1
    fmt.bitsPerSample = 16
    fmt.frameTimeUsec = 20000
    amp.createPort("amp", fmt)
    ep.audDevManager().getCaptureDevMedia().startTransmit(amp)
    amp.startTransmit(ep.audDevManager().getPlaybackDevMedia())

    time.sleep(3)
    del amp
    write("************* Endpoint started ok, now shutting down... *************" + "\r\n")
    ep.libDestroy()

#
# Tone generator
#
def ua_tonegen_test():
    write("UA tonegen test.." + "\r\n")
    ep_cfg = pj.EpConfig()

    ep = pj.Endpoint()
    ep.libCreate()
    ep.libInit(ep_cfg)
    ep.libStart()

    tonegen = pj.ToneGenerator()
    tonegen.createToneGenerator()

    tone = pj.ToneDesc()
    tone.freq1 = 400
    tone.freq2 = 600
    tone.on_msec = 1000
    tone.off_msec = 1000
    tones = pj.ToneDescVector()
    tones.append(tone)

    digit = pj.ToneDigit()
    digit.digit = '0'
    digit.on_msec = 1000
    digit.off_msec = 1000
    digits = pj.ToneDigitVector()
    digits.append(digit)

    adm = ep.audDevManager()
    spk = adm.getPlaybackDevMedia()

    tonegen.play(tones, True)
    tonegen.startTransmit(spk)
    time.sleep(5)

    tonegen.stop()
    tonegen.playDigits(digits, True)
    time.sleep(5)

    dm = tonegen.getDigitMap()
    write(dm[0].digit + "\r\n")
    dm[0].freq1 = 400
    dm[0].freq2 = 600
    tonegen.setDigitMap(dm)

    tonegen.stop()
    tonegen.playDigits(digits, True)
    time.sleep(5)

    tonegen = None

    ep.libDestroy()

class RandomIntVal():
    def __init__(self):
        self.value = randint(0, 100000)

class TestJob(pj.PendingJob):
    def __init__(self):
        super().__init__()
        self.val = RandomIntVal()
        print("Job created id:", id(self), "value:", self.val.value)

    def execute(self, is_pending):
        print("Executing job value:", self.val.value, is_pending)

    def __del__(self):
        print("Job deleted id:", id(self), "value:", self.val.value)

def add_new_job1(ep):
    print("Creating job 1")
    job = TestJob()
    print("Adding job 1")
    ep.utilAddPendingJob(job)

# Function to be executed in a separate thread
def add_new_job2(ep):
    ep.libRegisterThread("thread")
    print("Creating job 2")
    job = TestJob()
    print("Adding job 2")
    ep.utilAddPendingJob(job)

def ua_pending_job_test():
    write("PendingJob test.." + "\r\n")
    ep_cfg = pj.EpConfig()
    ep_cfg.uaConfig.threadCnt = 0
    ep_cfg.uaConfig.mainThreadOnly = True

    ep = pj.Endpoint()
    ep.libCreate()
    ep.libInit(ep_cfg)
    ep.libStart()

    # test 1
    # adding job from a separate function
    add_new_job1(ep)

    # test 2
    # adding job from a different thread
    my_thread = threading.Thread(target=add_new_job2, args=(ep,))
    my_thread.start()

    time.sleep(1)
    print("Collecting gc 1")
    gc.collect()

    print("Handling events")
    for _ in range(100):
        ep.libHandleEvents(10)

    print("Collecting gc 2")
    gc.collect()

    ep.libDestroy()

#
# main()
#
if __name__ == "__main__":
    ua_data_test()
    ua_run_test_exception()
    ua_run_log_test()
    ua_run_ua_test()
    ua_tonegen_test()
    ua_pending_job_test()
    sys.exit(0)





#
from inc_cfg import *

ADD_PARAM = ""

if (HAS_SND_DEV == 0):
	ADD_PARAM += "--null-audio"

# Call with L16/8000/1 codec
test_param = TestParam(
		"PESQ codec L16/8000/1",
		[
			InstanceParam("UA1", ADD_PARAM + " --max-calls=1 --add-codec L16/8000/1 --clock-rate 8000 --play-file wavs/input.8.wav"),
			InstanceParam("UA2", "--null-audio --max-calls=1 --add-codec L16/8000/1 --clock-rate 8000 --rec-file  wavs/tmp.8.wav --auto-answer 200")
		]
		)

pesq_threshold = 3.5



#
import inc_const as const

PJSUA = ["--null-audio --max-calls=1 --id sip:pjsua@localhost --add-buddy $SIPP_URI"]

PJSUA_EXPECTS = [[0, "", "s"],
		 [0, "Subscribe presence of:", "1"],
		 [0, "Presence subscription .* is TERMINATED", ""],
		 [0, "Resubscribing .* in 5000 ms", ""]
		 ]

PJSUA_CLI_EXPECTS = [[0, "", "im sub_pre 1"],
		 [0, "Presence subscription .* is TERMINATED", ""],
		 [0, "Resubscribing .* in 5000 ms", ""]
		 ]



#
# cfg_site_sample.py - Sample site configuration
#
# Copyright (C) 2008-2009 Teluu Inc. (http://www.teluu.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import builder

# Your site name
SITE_NAME="Newham3"

# The URL where tests will be submitted to
URL = "http://192.168.0.2/dash/submit.php?project=PJSIP"

# Test group
GROUP = "Experimental"

# PJSIP base directory
BASE_DIR = "/root/project/pjproject"

# List of additional ccdash options
#OPTIONS = ["-o", "out.xml", "-y"]
OPTIONS = []

# What's the content of config_site.h
CONFIG_SITE = ""

# What's the content of user.mak
USER_MAK = ""

# List of regular expression of test patterns to be excluded
EXCLUDE = []

# List of regular expression of test patterns to be included (even
# if they match EXCLUDE patterns)
NOT_EXCLUDE = []
#"configure", "update", "build.*make", "build", "run.py mod_run.*100_simple"]



import inc_sip as sip
import inc_sdp as sdp

pjsua = "--null-audio --id=sip:CLIENT --registrar sip:127.0.0.1:$PORT " + \
	"--realm=python --user=username --password=password " + \
	"--auto-update-nat=0 --reg-timeout 300"

# 423 Response with Min-Expires header that is lower than what the client
# had requested
req1 = sip.RecvfromTransaction("Initial request", 423,
				include=["REGISTER sip"], 
				exclude=[],
				resp_hdr=["Min-Expires: 250"],
				expect="invalid Min-Expires"

			  	)

recvfrom_cfg = sip.RecvfromCfg("Invalid 423 response to REGISTER",
			       pjsua, [req1])
